"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

import numpy as np
from scipy.stats import (multivariate_normal as mvn, norm)
from scipy.stats._multivariate import _squeeze_output
import scipy
from fmvmm.mixtures.skewnormmix_smsn import d_mixedmvSN, dmvSN
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix_skewnormal

def make_matrix_non_singular(sigma_matrix, epsilon=1e-6):
    """
    Ensure that a matrix is non-singular by adding a small diagonal perturbation.

    Parameters:
    - sigma_matrix: The input matrix to make non-singular.
    - epsilon: A small positive value to add to the diagonal elements.

    Returns:
    - A non-singular matrix.
    """
    # Check if the matrix is singular
    if np.linalg.matrix_rank(sigma_matrix) < min(sigma_matrix.shape):
        # The matrix is singular, let's make it non-singular
        sigma_matrix += np.eye(sigma_matrix.shape[0]) * epsilon

    return sigma_matrix

def _check_params(mu, sigma, lmbda):
    """
    Internal method to check if passed in parameters are valid.

    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: Tuple of the parameters as np.ndarrays, if no assertion error
        occurs.
    :rtype: tuple(np.ndarray, np.ndarray, np.ndarray)
    """
    mu = np.asarray(mu)
    sigma = np.asarray(sigma)
    lmbda = np.asarray(lmbda)

    # reshape mu, lmbda to (d,1)
    if len(mu.shape) == 1: mu = mu[:, None]
    if len(lmbda.shape) == 1: lmbda = lmbda[:, None]
    
    # assert shapes (d,1), (d,d), (d,1)
    d00, d01 = mu.shape
    d10, d11 = lmbda.shape
    d20, d21 = sigma.shape
    assert d00==d10==d20, "mu, sigma, lmbda must have same dimension d"
    assert d20==d21, "sigma must be qudratic with shape dxd"
    assert d01==1 and d11==1, "second dimension of mu, lmbda must be 1"
    
    return mu, sigma, lmbda

def mean(mu, sigma, lmbda):
    """
    Mean function of the multivariate skew normal distribution.

    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: The mean of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    mu, sigma, lmbda = _check_params(mu, sigma, lmbda)    

    val, vec = np.linalg.eig(sigma)
    sigma_half = vec @ np.diag(val**(1/2)) @ vec.T
    delta = lmbda / np.sqrt(1 + np.sum(lmbda**2))
    return mu + np.sqrt(2/np.pi) * sigma_half @ delta

def var(mu, sigma, lmbda):
    """
    Variance function of the multivariate skew normal distribution.

    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: The variance of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    mu, sigma, lmbda = _check_params(mu, sigma, lmbda)
    
    val, vec = np.linalg.eig(sigma)
    sigma_half = vec @ np.diag(val**(1/2)) @ vec.T
    delta = lmbda / np.sqrt(1 + np.sum(lmbda**2))
    return sigma - (2/np.pi) * sigma_half @ delta @ delta.T @ sigma_half

def logpdf(x, mu, sigma, lmbda):
    """
    Log-probability density function of the multivariate skew normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (p,).
    """
    mu, sigma, lmbda = _check_params(mu, sigma, lmbda)
    
    x    = mvn._process_quantiles(x, len(mu))
    pdf  = mvn(mu.flatten(), sigma,allow_singular=True).logpdf(x)
    sigma=make_matrix_non_singular(sigma)
    val, vec = np.linalg.eig(sigma)
    sigma_inv_half = vec @ np.diag(val**(-1/2)) @ vec.T
    a = (lmbda.T @ sigma_inv_half).flatten()
    cdf  = norm(0, 1).logcdf((x-mu.flatten()) @ a)
    return _squeeze_output(np.log(2) + pdf + cdf)

def pdf(x, mu, sigma, lmbda):
    """
    Probability density function of the multivariate skew normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: The density at each observation.
    :rtype: np.ndarray with shape (p,).
    """
    return np.exp(logpdf(x, mu, sigma, lmbda))

def loglike(x, mu, sigma, lmbda):
    """
    Log-likelihood function of the multivariate skew normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :return: The log-likelihood for given all observations and parameters.
    :rtype: float
    """
    return np.sum(logpdf(x, mu, sigma, lmbda))

def total_params(mu, sigma, lmbda):
    p = len(mu)
    
    return 2*p + (p*(p+1)/2)

def rvs(mu, sigma, lmbda, size=1):
    """
    Random number generator of the multivariate skew normal distribution.

    :param mu: The location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: The scale parameter. A positive semi-definite array with
        shape (p, p).
    :type sigma: np.ndarray
    :param lmbda: The skewness parameter with shape (p,).
    :type lmbda: np.ndarray
    :param size: The number of samples to draw. Defaults to 1.
    :type size: int, optional
    :return: The random p-variate numbers generated.
    :rtype: np.ndarray with shape (n, p).
    """
    mu, sigma, lmbda = _check_params(mu, sigma, lmbda)

    # switch to (Omega, alpha) parameterization
    cov = sigma
    val, vec = np.linalg.eig(cov)
    cov_half = vec @ np.diag(val**(-1/2)) @ vec.T
    shape = cov_half @ lmbda

    aCa      = shape.T @ cov @ shape
    delta    = (1 / np.sqrt(1 + aCa)) * cov @ shape
    cov_star = np.block([[np.ones(1),     delta.T],
                        [delta, cov]])
    x        = mvn(np.zeros(len(shape)+1), cov_star).rvs(size)
    if size == 1: x = x[None, :]
    x0, x1   = x[:, 0], x[:, 1:]
    inds     = x0 <= 0
    x1[inds] = -1 * x1[inds]
    return x1 + mu.T


def fit(x, maxiter = 100, ptol = 1e-6, ftol = 1e-10, eps=0.9, return_loglike = False):
    """
    Estimate the parameters of the multivariate skew normal distribution
    using an EM algorithm.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param maxiter: The maximum number of iterations to use in the EM algorithm.
        Defaults to 100.
    :type maxiter: int, optional
    :param ptol: The relative convergence criterion for the estimated 
        parameters. Defaults to 1e-6.
    :type ptol: float, optional
    :param ftol: The relative convergence criterion for the log-likelihood 
        function. Defaults to np.inf.
    :type ftol: float, optional
    :param eps: Initialization shrinkage of delta parameter. Defaults to 0.9.
    :type eps: float, optional
    :param return_loglike: Return a list of log-likelihood values at each iteration. 
        Defaults to False.
    :type return_loglike: np.ndarray, optional
    :return: The fitted parameters (<array> mu, <array> scale, <array> nu). Also returns
        a list of log-likelihood values at each iteration of the EM algorithm if 
        ``return_loglike=True``.
    :rtype: tuple
    """

    def _ll(xi, n, omega_inv_half, delta):
    
        d = xi.shape[1]
        lmbda = delta / np.sqrt(1 - np.sum(delta**2))
        wx = xi @ omega_inv_half
        cdf = scipy.stats.norm.cdf(wx @ lmbda)
        pdf = scipy.stats.multivariate_normal.pdf(wx, np.zeros(d), np.eye(d))
        ww = n * np.log(2) + np.sum(np.log(cdf)) + np.sum(np.log(pdf)) + n*np.log(np.linalg.det(omega_inv_half))
        return ww
    
    itereps = 10e-6
    n, p = x.shape
    
    # get pseudo moment estiators???
    m1 = np.mean(x, axis=0)
    center = x - m1
    m3 = np.mean(center**3, axis=0)
    a1 = np.sqrt(2/np.pi)
    b1 = (4/np.pi-1)*a1
    gam = np.sign(m3) * (np.abs(m3)/b1)**(1/3)

    mu0 = m1 - a1 * gam
    omega0 = (center.T @ center) / n + a1 * (gam[:, None] @ gam[None, :])

    val, vec = np.linalg.eig(omega0)
    omega0_inv_half = vec @ np.diag(val**(-1/2)) @ vec.T
    omega0_half = vec @ np.diag(val**(1/2)) @ vec.T

    delta0 = omega0_inv_half @ gam
    
    
    if (np.sum(delta0**2) > 1):
        delta0 = eps * delta0 / np.sqrt(np.sum(delta0**2))

    tau0 = 1
    mu1 = mu0
    omega1 = omega0
    delta1 = delta0
    
    ll_val = np.array([_ll(x - mu0, n, omega0_inv_half, delta0)])
    
    for i in range(maxiter):
        
        # E-step
        xi0 = x - mu0
        lmbda0 = delta0 / np.sqrt(1 - np.sum(delta0**2))
        w1 = 1 / np.sqrt(1 + np.sum(lmbda0**2))
        wa = xi0 @ (omega0_inv_half @ lmbda0)
        pdf = scipy.stats.norm.pdf(wa)
        cdf = scipy.stats.norm.cdf(wa)

        yexabs = tau0 * w1 * (pdf/cdf + wa)
        yex2   = tau0**2 * w1**2 * wa * (pdf/cdf + wa + 1/wa)
        
        # M-step
        mu1 = m1 - (omega0_half @ delta0)/tau0 * np.mean(yexabs)
        tau1 = np.mean(yex2)**(1/2)
        omega1 = (x-mu1).T @ (x-mu1) / n
        val, vec = np.linalg.eig(omega1)
        omega1_inv_half = vec @ np.diag(val**(-1/2)) @ vec.T
        omega1_half = vec @ np.diag(val**(1/2)) @ vec.T        
        delta1 = omega1_inv_half @ (yexabs @ (x-mu1))/(n*tau1)
        

        # measure log likelihood
        ll_val = np.append(ll_val, _ll(x-mu1, n, omega1_inv_half, delta1))

        # check for converagence (stopping criterion)
        #have_params_converged = \
        #    np.all(np.abs(mu1 - mu0) <= .5 * ptol * (abs(mu1) + abs(mu0))) & \
        #    np.all(np.abs(omega1 - omega0) <= .5 * ptol * (abs(omega1) + abs(omega0))) & \
        #    np.all(np.abs(delta1 - delta0) <= .5 * ptol * (abs(delta1) + abs(delta0)))
        has_fun_converged = np.abs(ll_val[-1] - ll_val[-2]) <= .5 * ftol * ( np.abs(ll_val[-1]) + np.abs(ll_val[-2]))
        if has_fun_converged: break

        # update old parameters
        mu0 = mu1
        omega0_inv_half = omega1_inv_half
        omega0_half = omega1_half
        delta0 = delta1
        tau0 = tau1

    #lmbda1 = delta0 / np.sqrt(1-np.sum(delta0**2))
    #val, vec = np.linalg.eig(omega1)
    #inv_half = vec @ np.diag(val**(-1)) @ vec.T
    #lmbda0 = inv_half @ delta1 / np.sqrt(1 - np.sum(delta1**2))
    lmbda0 = delta1 / np.sqrt(1 - np.sum(delta1**2))
    if return_loglike:
        return mu1, omega1, lmbda0, ll_val
    return mu1, omega1, lmbda0

def info_mat(X,mu,sigma,lmbda):
    
    p =len(mu)
    g = 1
    
    IM = info_matrix_skewnormal(
    X, [1], [mu], [sigma], [lmbda],
    d_mixedmvSN_func=d_mixedmvSN, 
    dmvSN_func=dmvSN, 
    g=g, 
    p=p
)
    
    final_IM = expand_reduced_IM_to_full_no_pi(IM, p, g)
    
    return final_IM

def expand_reduced_IM_to_full_no_pi(IM_reduced, p, g):
    """
    Expand and remove mixture-weight parameters from the 'reduced' info matrix
    of a Skew-Normal mixture, as produced by info_matrix_skewnormal(...).

    In info_matrix_skewnormal, each cluster j has:
      - mu_j : p
      - shape_j : p
      - Sigma_j : p*(p+1)//2   (upper-triangular portion)
    => cluster_params_reduced = 2*p + (p*(p+1)//2)
    => total dimension:  g*cluster_params_reduced + (g-1 if g>1 else 0)

    Steps:
      1) Remove (g-1) rows/cols for pi_j if g>1.
      2) Expand each Sigma_j from p*(p+1)//2 to p^2.
      3) Return the expanded info matrix with dimension
         g*(2p + p^2), i.e. only mu_j, shape_j, full Sigma_j.

    Parameters
    ----------
    IM_reduced : ndarray of shape (M_reduced, M_reduced)
        The reduced info matrix from info_matrix_skewnormal(...).
    p : int
        Dimension of each component (size of mu_j, shape_j).
    g : int
        Number of mixture components.

    Returns
    -------
    IM_full_no_pi : ndarray of shape (M_full_no_pi, M_full_no_pi)
        The expanded info matrix, with no pi_j parameters and
        full (p^2) blocks for each Sigma_j.
    """
    # 1) Dimensions in the "reduced" format
    cluster_params_reduced = 2*p + (p*(p+1)//2)
    M_reduced_total = g*cluster_params_reduced + (g-1 if g>1 else 0)  # no extra 'nu' in pure Skew-Normal
    if IM_reduced.shape != (M_reduced_total, M_reduced_total):
        raise ValueError(
            f"Expected IM_reduced to be {M_reduced_total} x {M_reduced_total}, "
            f"but got {IM_reduced.shape}."
        )

    # 2) After removing pi block (if g>1), we have:
    M_reduced_no_pi = g*cluster_params_reduced

    # 3) After expanding Sigma to p^2 for each cluster, final dimension is:
    #    M_full_no_pi = g*(2p + p^2).
    M_full_no_pi = g * (2*p + p*p)

    # --- Remove the pi block if g>1
    if g > 1:
        offset_pi = g*cluster_params_reduced  # start of pi block
        keep_mask = np.ones(M_reduced_total, dtype=bool)
        # The pi parameters are offset_pi : offset_pi + (g-1)
        pi_indices = np.arange(offset_pi, offset_pi + (g-1))
        keep_mask[pi_indices] = False
        IM_reduced_no_pi = IM_reduced[keep_mask][:, keep_mask]
    else:
        # If g=1, no pi
        IM_reduced_no_pi = IM_reduced

    if IM_reduced_no_pi.shape != (M_reduced_no_pi, M_reduced_no_pi):
        raise ValueError(
            f"After removing pi, expected shape ({M_reduced_no_pi}, {M_reduced_no_pi}), "
            f"but got {IM_reduced_no_pi.shape}."
        )

    # --- 4) Build duplication matrix D_no_pi
    # shape: (M_full_no_pi, M_reduced_no_pi)
    D_no_pi = np.zeros((M_full_no_pi, M_reduced_no_pi), dtype=float)

    offset_reduced = 0
    offset_full = 0

    def copy_block_1to1(num_params):
        # copies 'num_params' parameters in a 1-to-1 fashion
        nonlocal offset_reduced, offset_full
        for kk in range(num_params):
            D_no_pi[offset_full + kk, offset_reduced + kk] = 1.0
        offset_reduced += num_params
        offset_full    += num_params

    # For each cluster j => we have mu_j (p), shape_j (p), Sigma_j (p*(p+1)//2)
    for j_ in range(g):
        # mu_j => p
        copy_block_1to1(p)
        # shape_j => p
        copy_block_1to1(p)

        # Expand Sigma_j from p*(p+1)//2 => p^2
        old_start = offset_reduced
        new_start = offset_full
        old_index = 0

        for row_ in range(p):
            for col_ in range(row_, p):
                old_idx = old_start + old_index
                old_index += 1
                new_idx1 = new_start + (row_*p + col_)
                new_idx2 = new_start + (col_*p + row_)

                D_no_pi[new_idx1, old_idx] = 1.0
                if new_idx2 != new_idx1:
                    D_no_pi[new_idx2, old_idx] = 1.0

        offset_reduced += (p*(p+1)//2)
        offset_full    += (p*p)

    # --- 5) Final expanded info matrix
    IM_full_no_pi = D_no_pi @ IM_reduced_no_pi @ D_no_pi.T
    return IM_full_no_pi



import numpy as np
from fmvmm.mixtures import skewtmix_smsn
from fmvmm.mixsmsn.dens import dmvt_ls, d_mixedmvST
from fmvmm.mixsmsn.gen import gen_ST_multi
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix_skewt

def logpdf(x,mu, sigma, lmbda, nu):
    
    return np.log(dmvt_ls(x,mu, sigma, lmbda, nu))

def pdf(x,mu, sigma, lmbda, nu):
    
    return dmvt_ls(x,mu, sigma, lmbda, nu)

def loglike(x,mu, sigma, lmbda, nu):
    
    return np.sum(logpdf(x, mu, sigma, lmbda, nu))

def total_params(mu, sigma, lmbda, nu):
    p = len(mu)
    
    return 2*p + (p*(p+1)/2) +1

def rvs(mu, sigma, lmbda, nu, size = 1):
    
    return gen_ST_multi(size, mu, sigma, lmbda, nu)

def fit(x):
    model = skewtmix_smsn.SkewTMix(1, verbose = False)
    model.fit(x)
    _, alphas = model.get_params()
    
    return alphas[0][0], alphas[0][1], alphas[0][2], np.array([alphas[0][3]])

def info_mat(X,mu,sigma,lmbda,nu):
    
    p =len(mu)
    g = 1
    if isinstance(nu, np.ndarray):
        nu = nu[0]
    IM = info_matrix_skewt(
    X,[1], [mu], [sigma], [lmbda], nu,
    d_mixedmvST_func=d_mixedmvST,
    dmvt_ls_func=dmvt_ls,   
    g=g, p=p)
    
    final_IM = expand_reduced_IM_to_full_no_pi(IM, p, g)
    
    return final_IM
    

def expand_reduced_IM_to_full_no_pi(IM_reduced, p, g):
    """
    Take a 'reduced' information matrix (including mu_j, shape_j, upper-triangular
    Sigma_j, possibly also mixture weights pi_j, and nu), then:

      1) Remove all rows/columns corresponding to pi_j (the mixture weights).
         (If g=1, there are no such rows/columns to remove.)
      2) Expand each covariance block from size p(p+1)/2 to p^2 by duplicating
         symmetric entries.
      3) Return the resulting full information matrix that has mu_j, shape_j,
         *full* Sigma_j, and nu — no mixture weights.

    Parameter layout in IM_reduced:
      - For j=1..g:
        mu_j (p), shape_j (p), Sigma_j (p*(p+1)//2)
      - If g>1 => plus (g-1) for pi
      - Finally +1 for nu
    That is:
      M_reduced_total = g*(2p + p(p+1)//2) + (g-1) + 1

    We remove the pi rows/columns, so dimension becomes:
      M_reduced_no_pi = g*(2p + p(p+1)//2) + 1

    Then we expand Sigma blocks to p^2:
      M_full_no_pi = g*(2p + p^2) + 1

    Returns
    -------
    IM_full_no_pi : (M_full_no_pi, M_full_no_pi) ndarray
    """

    # -------------------------------
    # 1) Basic dimension checks
    cluster_params_reduced = 2*p + (p*(p+1)//2)
    M_reduced_total = g*cluster_params_reduced + (g-1) + 1  # total in old matrix
    if IM_reduced.shape != (M_reduced_total, M_reduced_total):
        raise ValueError(
            f"Expected IM_reduced to be {M_reduced_total}x{M_reduced_total}, "
            f"but got {IM_reduced.shape}."
        )

    # The dimension after removing pi:
    M_reduced_no_pi = g*cluster_params_reduced + 1  # no (g-1)
    M_full_no_pi    = g*(2*p + p*p) + 1

    # -------------------------------
    # 2) Remove pi rows/columns if g>1
    #    pi_j occupy the block right after all cluster blocks:
    #    offset_pi = g*cluster_params_reduced, size = (g-1)
    offset_pi = g*cluster_params_reduced
    if g > 1:
        # Build an index mask for the rows/cols we want to keep:
        keep_mask = np.ones(M_reduced_total, dtype=bool)
        # Turn off the pi indices
        pi_indices = np.arange(offset_pi, offset_pi+(g-1), dtype=int)
        keep_mask[pi_indices] = False
        # keep all others => we also keep the last param for nu
        IM_reduced_no_pi = IM_reduced[keep_mask][:, keep_mask]
    else:
        # If g=1, there are no pi parameters
        IM_reduced_no_pi = IM_reduced

    # sanity check
    if IM_reduced_no_pi.shape != (M_reduced_no_pi, M_reduced_no_pi):
        raise ValueError(
            f"After removing pi, expect shape = {M_reduced_no_pi}x{M_reduced_no_pi} "
            f"but got {IM_reduced_no_pi.shape}"
        )

    # -------------------------------
    # 3) Build the duplication matrix D_no_pi
    #    which maps from "reduced no-pi" coords to "full no-pi" coords.
    D_no_pi = np.zeros((M_full_no_pi, M_reduced_no_pi), dtype=float)

    offset_reduced = 0
    offset_full    = 0

    def copy_block_1to1(n_params):
        """
        Write a diagonal identity sub-block of size n_params
        from offset_reduced -> offset_full.
        """
        nonlocal offset_reduced, offset_full
        for k_ in range(n_params):
            D_no_pi[offset_full + k_, offset_reduced + k_] = 1.0
        offset_reduced += n_params
        offset_full    += n_params

    # For each cluster j, we have: mu_j (p), shape_j(p), Sigma_j( p*(p+1)//2 )
    for j_ in range(g):
        # 3.1) mu_j => p
        copy_block_1to1(p)
        # 3.2) shape_j => p
        copy_block_1to1(p)

        # 3.3) expand Sigma from p*(p+1)//2 to p^2
        old_start = offset_reduced
        new_start = offset_full
        old_index = 0
        for r_ in range(p):
            for c_ in range(r_, p):
                # old param index
                old_idx = old_start + old_index
                old_index += 1

                # new param indices (two if r_ != c_)
                new_idx1 = new_start + (r_*p + c_)
                new_idx2 = new_start + (c_*p + r_)

                D_no_pi[new_idx1, old_idx] = 1.0
                if new_idx2 != new_idx1:
                    D_no_pi[new_idx2, old_idx] = 1.0

        offset_reduced += (p*(p+1)//2)
        offset_full    += (p*p)

    # We have removed pi, so no block for them.

    # 3.4) final param = nu => 1 param
    copy_block_1to1(1)

    # -------------------------------
    # 4) Multiply to get final result
    IM_full_no_pi = D_no_pi @ IM_reduced_no_pi @ D_no_pi.T
    return IM_full_no_pi

    


"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

import numpy as np
import scipy.stats
from scipy.special import gammaln, kv, digamma
from fmvmm.distributions.gig import _besselM3, _check_gig_pars
from fmvmm.distributions import gig
import warnings
import math

def custom_log(input_value):
    if input_value == 0:
        return np.log1p(input_value)
    else:
        return np.log(input_value)

def process_variable(var):
    if np.isnan(var) or var == 0:
        return 0.1
    else:
        return var



def logpdf(x, lmbda, chi, psi, mu, sigma, gamma):
    """
    Log-probability density function of the generalised hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (n,).
    """

    ## Density of a multivariate generalized hyperbolic distribution.
    ## Covers all special cases as well.

    n, d = x.shape
    diff = x - mu
    det_sigma = np.linalg.det(sigma)
    inv_sigma = np.linalg.inv(sigma)

    Q = np.sum((diff @ inv_sigma) * diff, axis=1) # mahalanobis
    if np.sum(np.abs(gamma)) == 0:
        symm = True
        skewness_scaled = 0
        skewness_norm = 0
    else:
        symm = False
        skewness_scaled = (diff @ inv_sigma) @ gamma
        skewness_norm = (gamma @ inv_sigma) @ gamma
    out = np.nan

    if psi==0:
        lmbda_min_d_2 = lmbda - d/2
        if symm:
            # symmetric student's t
            interm = chi + Q

            log_const_top = -lmbda * np.log(chi) + gammaln(-lmbda_min_d_2)
            
            log_const_bottom = d/2 * np.log(np.pi) + 0.5 * np.log(det_sigma) + gammaln(-lmbda)
            
            log_top = lmbda_min_d_2 * np.log(interm)
            

            out = log_const_top + log_top - log_const_bottom

        else:
            # Asymmetric student's t
            
            interm = np.sqrt((chi + Q) * skewness_norm)

            log_const_top = -lmbda * np.log(chi) - lmbda_min_d_2 * np.log(skewness_norm)
            log_const_bottom = (d/2 * np.log(2 * np.pi) + 0.5 * np.log(det_sigma) + 
                gammaln(-lmbda) - (lmbda + 1) * np.log(2))
            log_top = _besselM3(lmbda_min_d_2, interm, logvalue = True) + skewness_scaled
            log_bottom = -lmbda_min_d_2 * np.log(interm)

            out = log_const_top + log_top - log_const_bottom - log_bottom

    elif psi > 0:
        lmbda_min_d_2 = lmbda - d/2.

        if chi > 0: # ghyp, hyp and NIG (symmetric and asymmetric)
            log_top = _besselM3(lmbda_min_d_2, np.sqrt((psi + skewness_norm) * (chi + Q)),
                                logvalue = True) + skewness_scaled
            log_bottom = -lmbda_min_d_2 * np.log(np.sqrt((psi + skewness_norm) * (chi + Q)))
            log_const_top = -lmbda/2 * np.log(psi * chi) + (d/2) * np.log(psi) - \
                lmbda_min_d_2 * np.log(1 + skewness_norm/psi)
            log_const_bottom = (d/2) * np.log(2*np.pi) + _besselM3(
                lmbda, np.sqrt(chi*psi), logvalue = True) + 0.5 * np.log(det_sigma)

            out = log_const_top + log_top - log_const_bottom - log_bottom

        elif chi == 0: # Variance gamma (symmetric and asymmetric)
            eps = 1e-8

            # Standardized observations that are close to 0 are set to 'eps'.
            if np.any(Q < eps):
                # If lambda == 0.5 * dimension, there is another singularity.
                if np.abs(lmbda_min_d_2) < eps:
                    raise Exception("Unhandled singularity: Some standardized observations are close to 0 (< " + \
                                    str(eps) + ") and lambda is close to 0.5 * dimension!\n")
                else:
                    Q[Q < eps] = eps
                    Q[Q == 0] = eps
                    warnings.warn("Singularity: Some standardized observations are close to 0 (< " + \
                                  str(eps) + ")!\nObservations set to " + str(eps) + ".\n")

            log_top = _besselM3(lmbda_min_d_2, np.sqrt((psi + skewness_norm) * (chi + Q)), logvalue=True) + skewness_scaled
            log_bottom = -lmbda_min_d_2 * np.log(np.sqrt((psi + skewness_norm) * (chi + Q)))
            log_const_top = d * np.log(psi)/2 + (1-lmbda) * np.log(2) - \
                lmbda_min_d_2 * np.log(1 + skewness_norm/psi)
            log_const_bottom = (d/2) * np.log(2 * np.pi) + gammaln(lmbda) + 0.5 * np.log(det_sigma)
            out = log_const_top + log_top - log_const_bottom - log_bottom
        else:
            out = np.nan

    return out

def pdf(x, lmbda, chi, psi, mu, sigma, gamma):
    """
    Probability density function of the generalised hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The density at each observation.
    :rtype: np.ndarray with shape (n,).
    """
    return np.exp(logpdf(x, lmbda, chi, psi, mu, sigma, gamma))

def loglike(x, lmbda, chi, psi, mu, sigma, gamma):
    """
    Log-likelihood function of the generalised hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The log-likelihood given all observations and parameters.
    :rtype: float
    """
    return np.sum(logpdf(x, chi, lmbda, psi, mu, sigma, gamma))


def total_params(lmbda, chi, psi, mu, sigma, gamma):
    p = len(mu)
    
    return 3 + 2*p + (p*(p+1)/2)


def rvs(lmbda, chi, psi, mu, sigma, gamma, size=1):
    """
    Random number generator of the generalised hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :param size: The number of samples to draw. Defaults to 1.
    :type size: int, optional
    :return: The random p-variate numbers generated.
    :rtype: np.ndarray with shape (n, p).
    """
    d = len(mu)
    Z = np.random.normal(size=(size, d))
    A = np.linalg.cholesky(sigma).T
    W = gig.rvs(lmbda, chi, psi, size)
    
    return np.sqrt(W)[:, np.newaxis] * Z.dot(A) + np.tile(mu, (size, 1)) + np.outer(W, gamma)

def mean(lmbda, chi, psi, mu, sigma, gamma):
    """
    Mean function of the generalised hyperbolic distribution. We use
    the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The mean of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    if psi==0: # gig -> invgamma
        if lmbda > -1:
            raise Exception("Mean for psi==0 is only defined for lambda < -1")
        alpha = -lmbda
        beta = 0.5*chi
        EW = beta/(alpha-1)
    elif psi>0:
        if chi==0: # gig -> gamma
            alpha = lmbda
            beta = 0.5*psi
            EW = beta/alpha
        elif chi>0: # gig
            alpha = np.sqrt(chi*psi)
            EW = np.sqrt(chi/psi) * kv(lmbda+1, alpha) / kv(lmbda, alpha)

    return mu + EW*gamma

def var(lmbda, chi, psi, mu, sigma, gamma):
    """
    Variance function of the generalised hyperbolic distribution. We use
    the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The variance of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    if psi==0: # gig -> invgamma
        if lmbda > -2:
            raise Exception("Var for psi==0 is only defined for lambda < -2")
        alpha = -lmbda
        beta = 0.5*chi
        EW = beta/(alpha-1)
        VarW = beta**2 / ((alpha-1)**2 * (alpha-2))
    elif psi>0:
        if chi==0: # gig -> gamma
            alpha = lmbda
            beta = 0.5*psi
            EW = beta/alpha
            VarW = alpha/(beta**2)
        elif chi>0: # gig
            alpha = np.sqrt(chi*psi)
            EW = np.sqrt(chi/psi) * kv(lmbda+1, alpha) / kv(lmbda, alpha)
            EW2 = (chi/psi) * kv(lmbda+2, alpha) / kv(lmbda, alpha)
            VarW = EW2 - EW**2

    return VarW * np.outer(gamma, gamma) + EW * sigma


# def _alphabar2chipsi(alpha_bar, lmbda):
#     """
#     From (alpha_bar, lambda) parameterisation to (chi, psi)
#     """
#     #k1=process_variable(kv(lmbda+1, alpha_bar))
#     #k2=process_variable(kv(lmbda, alpha_bar))

#     #psi = process_variable(alpha_bar) * k1 / k2
#     if alpha_bar==0 and lmbda >0:
#         #VG Case
#         chi = 0
#         psi = 2*lmbda
#     elif alpha_bar==0 and lmbda <0:
#         #Student t case
#         chi = -2*(lmbda+1)
#         psi = 0
#     else:
#         psi = alpha_bar * kv(lmbda+1, alpha_bar) / kv(lmbda, alpha_bar)
#         chi = alpha_bar**2 / psi
#     #if psi==0:
#         #chi=alpha_bar**2/0.5
#     #else:
#         #chi = alpha_bar**2 / psi
#     return chi, psi

def _alphabar2chipsi(alpha_bar, lmbda, eps=1e-15):
    """
    Translate (alpha_bar, lambda) -> (chi, psi) as in R's .abar2chipsi(),
    with clamping to avoid Inf/NaN if the ratio overflows.
    """
    if alpha_bar < 0:
        raise ValueError("alpha_bar must be non-negative.")

    # --- Special cases alpha_bar ~ 0 ---
    if alpha_bar < eps:
        if lmbda > 0:     # VG
            chi = 0.0
            psi = 2*lmbda
        elif lmbda < 0:   # Student-t
            psi = 0.0
            chi = -2*(lmbda + 1)
        else:
            raise ValueError("Forbidden combination: alpha_bar=0 and lambda=0")
        return chi, psi

    # --- alpha_bar > 0 case ---
    # Use the same branching as the R code:
    if lmbda >= 0:
        #   psi = alpha_bar * K_{lambda+1} / K_{lambda}
        #   chi = alpha_bar^2 / psi
        denom = kv(lmbda,   alpha_bar)
        numer = kv(lmbda+1, alpha_bar)
        # If the ratio is 0 or inf, handle gracefully:
        if denom == 0 or not np.isfinite(denom):
            # R code sets psi=200 on besselK overflow
            psi = 200.0
        else:
            psi = alpha_bar * numer / denom
            if not np.isfinite(psi):
                psi = 200.0
        chi = alpha_bar**2 / psi

    else:
        #   chi = alpha_bar * K_{lambda} / K_{lambda+1}
        #   psi = alpha_bar^2 / chi
        denom = kv(lmbda+1, alpha_bar)
        numer = kv(lmbda,   alpha_bar)
        if denom == 0 or not np.isfinite(denom):
            chi = 200.0
        else:
            chi = float(alpha_bar * numer / denom)
            if not np.isfinite(chi):
                chi = 200.0
        psi = alpha_bar**2 / chi

    return float(chi), float(psi)


def t_optfunc(thepars, delta_sum, xi_sum, n_rows):
    """
    Log-likelihood function of the inverse gamma distribution
    """
    nu = -2 * (-1 - np.exp(thepars))
    term1 = -n_rows * nu * np.log(nu/2 - 1)/2
    term2 = (nu/2 + 1) * xi_sum + (nu/2 - 1) * delta_sum
    term3 = n_rows * gammaln(nu/2)
    out = term1 + term2 + term3
    return out

# loglikelihood function of the gamma distribution
def vg_optfunc(thepars, xi_sum, eta_sum, n_rows):
    """
    Log-likelihood function of the gamma distribution
    """
    # print("doing vg")
    thepars = np.exp(thepars)
    term1 = n_rows * (thepars * np.log(thepars) - gammaln(thepars))
    term2 = (thepars - 1) * xi_sum - thepars * eta_sum
    out = -(term1 + term2)
    return out



def gig_optfunc(thepars, mix_pars_fixed, pars_order, delta_sum, eta_sum, xi_sum, n_rows):
    """
    Log-likelihood function of the generalized inverse gaussian distribution
    """
    # print("thepars",thepars)
    # print("mix_pars_fixed",mix_pars_fixed)
    out = np.nan
    tmp_pars = np.concatenate([thepars, mix_pars_fixed])
    # print("tmp_pars",tmp_pars)
    lmbda = tmp_pars[pars_order=="lmbda"]
    # print("lmbda",lmbda)
    alpha_bar = np.exp(tmp_pars[pars_order=="alpha_bar"])
    # print("alpha_bar",alpha_bar)
    chi, psi = _alphabar2chipsi(alpha_bar, lmbda)
    # print("chi, psi",chi, psi)
    if lmbda < 0 and psi == 0: # t
        out = t_optfunc(lmbda, delta_sum, xi_sum, n_rows)
    elif lmbda > 0 and chi == 0: # VG
        out = vg_optfunc(lmbda, xi_sum, eta_sum, n_rows)
    else: # ghyp, hyp, NIG
        term1 = (lmbda - 1) * xi_sum
        term2 = -chi * delta_sum/2
        term3 = -psi * eta_sum/2
        term4 = -n_rows * lmbda * np.log(chi)/2 + n_rows * lmbda * np.log(psi)/2 - \
            n_rows * _besselM3(lmbda, np.sqrt(chi * psi), logvalue = True)
        out = -(term1 + term2 + term3 + term4)
    return out

def _check_norm_pars(mu, sigma, gamma, dimension):
    """
    Check normal and skewness parameters  for consistency
    """
    if len(mu) != dimension:
        raise Exception("Parameter 'mu' must be of length " + str(dimension) + "!")
    if len(gamma) != dimension:
        raise Exception("Parameter 'gamma' must be of length " + str(dimension) + "!")
    if dimension > 1: # MULTIVARIATE
        if len(sigma.shape) != 2 or sigma.shape[0] != dimension or sigma.shape[1] != dimension:
            raise Exception("'sigma' must be a quadratic matrix with dimension " +\
                 str(dimension) + " x " + str(dimension) + "!")
    else: # UNIVARIATE
        if len(sigma) != dimension:
            raise Exception("Parameter 'sigma' must be a scalar!")

def fitghypmv(
    x, lmbda = 1, alpha_bar = 1, mu = None, sigma = None, gamma = None, symmetric = False,
    standardize = False, nit = 2000, reltol = 1e-8, abstol = 1e-7, silent = False,
    opt_pars = {"lmbda": True, "alpha_bar": True, "mu": True, "sigma": True, "gamma": True}
):
    """
    Estimate the parameters of the generalised hyperbolic distribution. We use
    the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation. Covers all special
    cases as well.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param lmbda: The intial value of lmbda. Defaults to 1.
    :tpe lmbda: float, optional
    :param alpha_bar: The initial value of alpha_bar. Defaults to 1.
    :type alpha_bar: float, optional
    :param mu: Optional initial value of mu, an array of shape (p,)
    :type mu: np.ndarray, optional
    :param sigma: Optional initial value of sigma, an array of shape (p,p)
    :type sigma: np.ndarray, optional
    :param gamma: Optional initial value of gamma, an array of shape (p,)
    :type gamma: np.ndarray, optional
    :param symmetric: Whether to fit a symmetric distribution or not. Default
        to False.
    :type symmetric: bool, optional
    :param standardize: Whether to standardize the data before fitting or not.
        Default to False.
    :type standardize: bool, optional
    :param nit: The maximum number of iterations to use in the EM algorithm.
        Defaults to 2000.
    :type nit: int, optional
    :param reltol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-8.
    :type reltol: float, optional
    :param abstol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-7.
    :type abstol: float, optional
    :param silent: Whether to print the log-likelihoods and parameter estimates
        during fitting or not. Defaults to False.
    :type silent: bool, optional
    :param opt_pars: A dict of (lmbda, alpha_bar, mu, sigma, gamma) with boolean values,
        denoting if the parameters should be estimated or fixed to their initial value.
        Defaults to fitting all parameters.
    :type opt_pars: dict, optional
    :return: The fitted parameters (<float> lmbda, <float> alpha_bar, <array> mu, <array> sigma,
        <array> gamma) and a list of log-likelihood values at each iteration of the EM algorithm,
        the number of performed iterations, whether the algorithm coverged, and the final AIC.
    :rtype: dict
    """

    n, d = x.shape
    chi, psi = _alphabar2chipsi(alpha_bar, lmbda)
    

    # check parameters of the mixing distribution for consistency
    _check_gig_pars(lmbda, chi, psi)

    # .check.opt.pars(opt.pars, symmetric)
    #opt.pars <- .check.opt.pars(opt.pars, symmetric)

    m1 = np.mean(x, axis=0)
    # center = x-m1
    center = m1 -x

    if mu is None: mu = m1
    if (gamma is None) or symmetric: gamma = np.zeros(d)
    if symmetric: opt_pars["gamma"] = False
    if sigma is None: sigma = np.cov(x.T)

    # check normal and skewness parameters  for consistency
    _check_norm_pars(mu, sigma, gamma, d)

    if standardize:
        # data will be standardized and initial values will be adapted
        tmp_mean = np.mean(x, axis=0)
        sigma_chol = np.linalg.cholesky(np.linalg.inv(np.cov(x.T))).T
        x = x - tmp_mean
        x = x @ sigma_chol
        sigma = sigma_chol.T @ sigma @ sigma_chol
        gamma = sigma_chol @ gamma
        mu = sigma_chol @ (mu - tmp_mean)

    # Initialize fitting loop
    i = 0
    rel_closeness = 100
    abs_closeness = 100
    tmp_fit = {"convergence": 0, "message": None}
    chi, psi = _alphabar2chipsi(alpha_bar, lmbda)
    # print("chi,psi:",chi,psi)
    ll = np.sum(logpdf(x, lmbda=lmbda, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma))
    # print("ll",ll)

    ll_save = []
    lmbda_save_save = []
    chi_save = []
    psi_save = []
    mu_save = []
    sigma_save = []
    gamma_save = []

    ## Start interations
    while (abs_closeness > abstol) and (rel_closeness > reltol) and (i < nit):
        i = i + 1

        ##<------------------------ E-Step: EM update ------------------------------>
        ## The parameters mu, sigma and gamma become updated

        inv_sigma = np.linalg.inv(sigma)

        diff = x-mu
        Q = np.sum((diff @ inv_sigma.T) * diff, axis=1)
        
        # Offset = (gamma @ inv_sigma) @ gamma
        Offset = gamma @ inv_sigma @ gamma.T
   
        delta = gig.expect(lmbda-d/2, Q+chi, psi+Offset, func = "1/x")
        delta_bar = np.mean(delta)
        eta = gig.expect(lmbda-d/2, Q+chi, psi+Offset, func = "x")
        eta_bar = np.mean(eta)

        if opt_pars["gamma"]:
            gamma = -(delta @ center) / (n*delta_bar*eta_bar - n)
            # print("gamma",gamma)

        if opt_pars["mu"]:
            mu = ((delta @ x)/n - gamma) / delta_bar
            # print("mu",mu)

        diff = x - mu
        tmp = delta[:, None] * diff
        
        if opt_pars["sigma"]:
            sigma = (tmp.T @ diff)/n - np.outer(gamma, gamma) * eta_bar
            # print("sigma",sigma)

        ##<------------------------ M-Step: EM update ------------------------------>
        ## Maximise the conditional likelihood function and estimate lambda, chi, psi

        inv_sigma = np.linalg.inv(sigma)
        Q = np.sum((diff @ inv_sigma) * diff, axis=1)
        
        Offset = gamma @ inv_sigma @ gamma.T
        
        xi_sum = np.sum(gig.expect(lmbda-d/2, Q+chi, psi+Offset, func="logx"))
        

        if alpha_bar==0 and lmbda > 0 and not opt_pars["alpha_bar"] and opt_pars["lmbda"]:
            ##<------  VG case  ------>
            
            eta_sum = np.sum(gig.expect(lmbda-d/2, Q+chi, psi+Offset, func="x"))
            
            minimizer_kwargs = {"method": "L-BFGS-B","args":(eta_sum, xi_sum, n)}
            
            x0= np.log(lmbda)
            
            tmp_fit = scipy.optimize.minimize(vg_optfunc, x0, args=(eta_sum, xi_sum, n), method='L-BFGS-B') #method='L-BFGS-B',np.log(lmbda)+2
            
            lmbda = np.exp(tmp_fit["x"])
            

        elif alpha_bar==0 and lmbda<0 and not opt_pars["alpha_bar"] and opt_pars["lmbda"]:
            ##<------  Student-t case  ------>
            delta_sum = np.sum(gig.expect(lmbda-d/2, Q+chi, psi+Offset, func="1/x"))
            
            x0= custom_log(-1 - lmbda) #np.log(lmbda) np.random.rand()
            tmp_fit = scipy.optimize.minimize(t_optfunc, x0, args=(delta_sum, xi_sum, n), method='L-BFGS-B') #custom_log(-1 - lmbda)+10
            
            lmbda = (-1 - np.exp(tmp_fit["x"]))
            
            

        elif opt_pars["lmbda"] or opt_pars["alpha_bar"]:
            ##<------  ghyp, hyp, NIG case  ------>
            delta_sum = np.sum(gig.expect(lmbda-d/2, Q+chi, psi+Offset, func="1/x"))
            eta_sum = np.sum(gig.expect(lmbda-d/2, Q+chi, psi+Offset, func="x"))

            mix_pars = {"lmbda": lmbda, "alpha_bar": np.log(alpha_bar)}
            thepars = {x: mix_pars[x] for x in ["lmbda", "alpha_bar"] if opt_pars[x]}
            mix_pars_fixed = {x: mix_pars[x] for x in ["lmbda", "alpha_bar"] if not opt_pars[x]}
            pars_order = np.array(list(thepars.keys()) + list(mix_pars_fixed.keys()))
            thepars_v = np.array(list(thepars.values())).flatten()
            mix_pars_fixed_v = np.array(list(mix_pars_fixed.values())).flatten()
            


            tmp_fit = scipy.optimize.minimize(gig_optfunc, thepars_v, args=(
                mix_pars_fixed_v, pars_order, delta_sum, eta_sum, xi_sum, n), method='L-BFGS-B')
            
            par = np.concatenate([tmp_fit["x"], mix_pars_fixed_v]) 


            lmbda = par[pars_order == "lmbda"]
            alpha_bar = np.exp(par[pars_order == "alpha_bar"])
            if np.isneginf(lmbda):
                lmbda=-50
            if np.isposinf(lmbda):
                lmbda=50
            if np.isneginf(alpha_bar):
                alpha_bar=-50
            if np.isposinf(alpha_bar):
                alpha_bar=50 
                


        chi, psi = _alphabar2chipsi(alpha_bar, lmbda)
        

        ## Test for convergence
        ll_old = ll
        ll = np.sum(logpdf(x, lmbda=lmbda, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma))

        abs_closeness = np.abs(ll - ll_old)
        rel_closeness = np.abs((ll - ll_old)/ll_old)

        if not np.isfinite(abs_closeness) or not np.isfinite(rel_closeness):
            warnings.warn("fit.ghypmv: Loglikelihood is not finite! Iteration stoped!\n" +\
                          "Loglikelihood :" + str(ll))
            break

        ll_save += [ll]
        lmbda_save_save += [lmbda]
        chi_save += [chi]
        psi_save += [psi]
        mu_save += [mu]
        sigma_save += [sigma]
        gamma_save += [gamma]

    ## END OF WHILE LOOP


    converged = False
    if i<nit and np.isfinite(rel_closeness) and np.isfinite(abs_closeness): converged = True
    # print("lmbda,alpha_bar",lmbda,alpha_bar)
    if(standardize):
        inv_sigma_chol = np.linalg.inv(sigma_chol)
        mu = inv_sigma_chol @ mu + tmp_mean
        sigma = inv_sigma_chol.T @ sigma @ inv_sigma_chol
        gamma = inv_sigma_chol @ gamma

        chi, psi = _alphabar2chipsi(alpha_bar, lmbda)
        ll = np.sum(logpdf(x, lmbda=lmbda, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma))


    nbr_fitted_params = opt_pars["alpha_bar"] + opt_pars["lmbda"] + d * (opt_pars["mu"]+opt_pars["gamma"]) +\
        (d/2)*(d+1)*opt_pars["sigma"]

    aic = -2 * ll + 2 * nbr_fitted_params

    return {"lmbda": lmbda, "alpha_bar": alpha_bar, "mu": mu, "sigma": sigma, "gamma": gamma,
            "ll": ll_save, "n_iter": i, "converged": converged, "aic": aic}



def fit(x, lmbda=1, alpha_bar=1, symmetric=False, standardize=False, nit=2000,
        reltol=1e-8, abstol=1e-7, silent=False, flambda=None, falpha_bar=None,
        fmu=None, fsigma=None, fgamma=None, return_loglike=False):
    """
    Estimate the parameters of the generalised hyperbolic distribution. We
    use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param lmbda: The intial value of lmbda. Defaults to 1.
    :tpe lmbda: float, optional
    :param alpha_bar: The initial value of alpha_bar. Defaults to 1.
    :type alpha_bar: float, optional
    :param symmetric: Whether to fit a symmetric distribution or not. Default
        to False.
    :type symmetric: bool, optional
    :param standardize: Whether to standardize the data before fitting or not.
        Default to False.
    :type standardize: bool, optional
    :param nit: The maximum number of iterations to use in the EM algorithm.
        Defaults to 2000.
    :type nit: int, optional
    :param reltol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-8.
    :type reltol: float, optional
    :param abstol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-7.
    :type abstol: float, optional
    :param silent: Whether to print the log-likelihoods and parameter estimates
        during fitting or not. Defaults to False.
    :type silent: bool, optional
    :param flmbda: If flmbda!=None, force lmbda to flmbda. Defaults to None.
    :type flmbda: np.ndarray, optional
    :param falpha_bar: If falpha_bar!=None, force alpha_bar to falpha_bar.
        Defaults to None.
    :type falpha_bar: np.ndarray, optional
    :param fmu: If fmu!=None, force mu to fmu. Defaults to None.
    :type fmu: np.ndarray, optional
    :param fsigma: If fsigma!=None, force sigma to fsigma. Defaults to None.
    :type fsigma: np.ndarray, optional
    :param fgamma: If fgamma!=None, force gamma to fgamma. Defaults to None.
    :type fgamma: np.ndarray, optional
    :param return_loglike: Return a list of log-likelihood values at each iteration.
        Defaults to False.
    :type return_loglike: np.ndarray, optional
    :return: The fitted parameters (<float> lmbda, <float> chi, <float> psi, <array> mu,
        <array> sigma, <array> gamma). Also returns a list of log-likelihood values at
        each iteration of the EM algorithm if ``return_loglike=True``.
    :rtype: tuple
    """
    
    opt_pars = {"lmbda": True, "alpha_bar": True,
                "mu": True, "sigma": True, "gamma": True}

    lmbda = lmbda if flambda is None else flambda
    alpha_bar = alpha_bar if falpha_bar is None else falpha_bar

    fit = fitghypmv(
        x, lmbda=lmbda, alpha_bar=alpha_bar, mu=fmu, sigma=fsigma, gamma=fgamma,
        symmetric=symmetric, standardize=standardize, nit=nit, reltol=reltol,
        abstol=abstol, silent=silent, opt_pars=opt_pars)

    if fit["alpha_bar"] != 0:
        chi, psi = _alphabar2chipsi(fit["alpha_bar"], fit["lmbda"])
    elif fit["alpha_bar"]==0 and fit["lmbda"] > 0:
         chi, psi = 0, 2 * fit["lmbda"]
    elif fit["alpha_bar"]==0 and fit["lmbda"] < 0:
         chi, psi = -2 * (fit["lmbda"] + 1), 0
    
    
    if return_loglike:
        return fit["lmbda"], chi, psi, fit["mu"], fit["sigma"], fit["gamma"], fit["ll"]
    return fit["lmbda"], chi, psi, fit["mu"], fit["sigma"], fit["gamma"]


def info_mat(x, lmbda, chi, psi, mu, sigma, gamma):
    from fmvmm.utils.utils_fmm import compute_info_scipy_fmvmm
    
    IM = compute_info_scipy_fmvmm(logpdf,[lmbda, np.array([chi]), np.array([psi]), mu, sigma, gamma],x)
    
    return IM


import numpy as np
import fmvmm.utils.dirichlet as drm

def logpdf(x,alpha):
    
    return np.log(drm.pdf(np.array(alpha))(x))

def pdf(x,alpha):
    
    return drm.pdf(np.array(alpha))(x)

def loglike(x,alpha):
    
    return np.sum(logpdf(x, np.array(alpha)))

def total_params(alpha):
    p = len(alpha)
    
    return p

def rvs(alpha, size = 1):
    
    return np.random.dirichlet(alpha,size)

def fit(x, method = "meanprecision"):
    
    return drm.mle(x, method = method)
    

    


import numpy as np
from fmvmm.mixtures._base import BaseMixture
from fmvmm.mixsmsn.dens import dmvt_ls, d_mixedmvST_custom
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from scipy.linalg import sqrtm, inv
import time
from scipy.optimize import minimize_scalar
from fmvmm.utils.utils_mixture import (mixture_clusters)
from scipy.special import gamma
from scipy.stats import t
from numpy.linalg import det
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix, standard_errors_from_info_matrix

def k_means_init_skewt(data, k):
    """
    Perform K-means initialization for a skew-t mixture model.

    Parameters:
    - data: pandas.DataFrame or np.ndarray of shape (n_samples, n_features).
    - k: Number of clusters.

    Returns:
    - pi_not: Mixing proportions for each cluster.
    - mu_not: Cluster means.
    - cov_not: Covariance matrices for each cluster.
    - shape_not: Shape (skewness) parameters for each cluster.
    - delta: Normalized skewness vectors for each cluster.
    - Delta: Skewness-adjusted covariance components for each cluster.
    - Gama: Residual covariance matrices for each cluster.
    """
    # Ensure data is a NumPy array
    if not isinstance(data, np.ndarray):
        data = data.to_numpy()

    n, p = data.shape

    # Perform K-means clustering
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)

    # Initialize cluster means
    mu_not = kmeans.cluster_centers_

    # Group data points by cluster
    data_cwise = [[] for _ in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data[i])
    
    # Convert grouped data to NumPy arrays
    data_cwise_arr = [np.asarray(cluster_data).T for cluster_data in data_cwise]

    # Initialize covariance matrices and skewness parameters
    alpha_not = []

    for j, cluster_data in enumerate(data_cwise_arr):
        # Covariance matrix
        cov_not_temp = np.cov(cluster_data)
        
        # Shape (skewness) parameter
        if cluster_data.shape[1] > 0:  # Ensure the cluster is not empty
            mean_diff = cluster_data.T - mu_not[j]
            skewness = np.sign(np.sum(mean_diff**3, axis=0))
    
        else:
            skewness = np.zeros(p)
        alpha_not.append((mu_not[j], cov_not_temp, skewness, 4))

    # Mixing proportions
    pi_not = [len(data_cwise[m]) / n for m in range(k)]

    return pi_not, alpha_not

def gmm_init_skewt(data, k):
    """
    Perform Gaussian Mixture Model (GMM) initialization for a skew-t mixture model.

    Parameters:
    - data: pandas.DataFrame or np.ndarray of shape (n_samples, n_features).
    - k: Number of clusters.

    Returns:
    - pi_not: Mixing proportions for each cluster.
    - mu_not: Cluster means.
    - cov_not: Covariance matrices for each cluster.
    - shape_not: Shape (skewness) parameters for each cluster.
    - delta: Normalized skewness vectors for each cluster.
    - Delta: Skewness-adjusted covariance components for each cluster.
    - Gama: Residual covariance matrices for each cluster.
    """
    # Ensure data is a NumPy array
    if not isinstance(data, np.ndarray):
        data = data.to_numpy()

    n, p = data.shape

    # Fit Gaussian Mixture Model
    gmm = GaussianMixture(n_components=k, random_state=1)
    gmm.fit(data)

    # Extract GMM parameters
    mu_not = gmm.means_  # Cluster means
    covariances = gmm.covariances_  # Covariance matrices
    if gmm.covariance_type == 'full':
        cov_not = covariances
    elif gmm.covariance_type == 'diag':
        cov_not = [np.diag(cov) for cov in covariances]
    elif gmm.covariance_type == 'tied':
        cov_not = [covariances for _ in range(k)]
    elif gmm.covariance_type == 'spherical':
        cov_not = [np.eye(p) * cov for cov in covariances]
    else:
        raise ValueError("Unsupported covariance type.")

    # Mixing proportions
    pi_not = gmm.weights_

    # Group data points by cluster
    labels = gmm.predict(data)
    data_cwise = [[] for _ in range(k)]
    for i in range(n):
        data_cwise[labels[i]].append(data[i])

    # Convert grouped data to NumPy arrays
    data_cwise_arr = [np.asarray(cluster_data).T for cluster_data in data_cwise]

    # Initialize skewness parameters and related components
    alpha_not = []
    shape_not = []
    delta = []
    Delta = []
    Gama = []

    for j, cluster_data in enumerate(data_cwise_arr):
        if cluster_data.shape[1] > 0:  # Ensure the cluster is not empty
            mean_diff = cluster_data.T - mu_not[j]
            skewness = np.sign(np.sum(mean_diff**3, axis=0))
            # shape_not.append(skewness)

            # Calculate delta, Delta, and Gama
            delta_j = skewness / np.sqrt(1 + skewness @ skewness)
            Delta_j = sqrtm(cov_not[j]) @ delta_j
            Gama_j = cov_not[j] - np.outer(Delta_j, Delta_j)

            delta.append(delta_j)
            Delta.append(Delta_j)
            Gama.append(Gama_j)
        else:
            # shape_not.append(np.zeros(p))
            skewness = np.zeros(p)
            delta.append(np.zeros(p))
            Delta.append(np.zeros(p))
            Gama.append(np.zeros((p, p)))
        alpha_not.append((mu_not[j], cov_not[j],skewness, 4))

    return pi_not, alpha_not


def estimate_alphas_skewt(
    X, gamma, alpha_prev, pi_prev,  # from the Soft EM (M-step)
    dmvt_ls_func = dmvt_ls, d_mixedmvST_func = d_mixedmvST_custom,
    bounds_nu=(1e-4, 100.0),
    tol_nu=1e-6
):
    """
    Single M-step update for a mixture of multivariate Skew-t components,
    following the logic from the R function smsn.mmix (Skew.t part).

    Parameters
    ----------
    X : ndarray of shape (n, p)
        Data matrix.
    gamma : ndarray of shape (n, g)
        Posterior responsibilities (soft cluster assignments).
    alpha_prev : list of length g
        Each element alpha_prev[j] = (mu_j, Sigma_j, shape_j, nu_j).
        We assume all clusters share the same nu, so alpha_prev[j][3] should be identical.
    pi_prev : 1D array of length g
        Mixing proportions from the previous iteration.
    dmvt_ls_func : callable
        Function dmvt_ls_func(X, mu, Sigma, shape, nu) -> PDF values (length n).
    d_mixedmvST_func : callable
        Function that computes the mixture PDF: sum_j pi_j * dmvt_ls(X,...).
    bounds_nu : tuple
        Bounds for optimizing nu in [low, high].
    tol_nu : float
        Tolerance for the nu optimizer.

    Returns
    -------
    alpha_new : list of length g
        Updated parameters for each cluster j:
        alpha_new[j] = (mu_j_new, Sigma_j_new, shape_j_new, nu_new).
        The same nu_new is placed in each cluster's 4th element.
    """

    n, p = X.shape
    g = len(alpha_prev)

    
    nu_prev = alpha_prev[0][3]

    # We will store intermediate items and final results here
    alpha_new = [None]*g

    # Precompute each cluster's old Delta, Gama, etc.
    # We do that so we can use "Delta_old" in the new mu update.
    Delta_old = []
    Gama_old = []
    for j in range(g):
        mu_j, Sigma_j, shape_j, _ = alpha_prev[j]
        # delta_j = shape_j / sqrt(1 + shape_j' shape_j)
        shape_j = np.asarray(shape_j).reshape(-1)  # ensure 1D
        denom = np.sqrt(1.0 + shape_j @ shape_j)
        delta_j = shape_j / denom  # p-vector

        # Delta_j = sqrtm(Sigma_j) @ delta_j
        Delta_j = sqrtm(Sigma_j) @ delta_j.reshape(-1,1)  # shape (p,1)

        # Gama_j = Sigma_j - Delta_j %*% t(Delta_j)
        Gama_j = Sigma_j - Delta_j@Delta_j.T

        Delta_old.append(Delta_j)
        Gama_old.append(Gama_j)

    

    def compute_E_u(X_i, mu_j, Sigma_j, Delta_j, Gama_j, shape_j, nu_val):
        """
        Compute the E[i], u[i] terms for a single point X_i in cluster j
        following the R code formula.
        """
        p = X_i.shape[0]
        dif_i = X_i - mu_j
        # For convenience, invert Gama_j once outside the loop in the M-step
        # but here we do a single case (we can pass a precomputed invGama_j if we want).
        return 0.0, 0.0  # Will fill in in the loop below

    alpha_new_list = []
    
    # Precompute cluster pdfs for each j to avoid repeated calls
    # We'll do it in a vectorized manner.
    cluster_pdf = np.zeros((n, g))
    for j in range(g):
        mu_j, Sigma_j, shape_j, nu_j = alpha_prev[j]
        cluster_pdf[:, j] = dmvt_ls_func(X, mu_j, Sigma_j, shape_j, nu_prev)

    # Mixture PDF for each i
    d_mixed = np.sum(pi_prev * cluster_pdf, axis=1)
    # Protect against zeros
    d_mixed[d_mixed < 1e-300] = 1e-300

    # Precompute inverse Sigma_j, inverse Gama_j
    inv_Sigma = []
    inv_Gama_old = []
    for j in range(g):
        _, Sigma_j, _, _ = alpha_prev[j]
        inv_Sigma.append(inv(Sigma_j))
        inv_Gama_old.append(inv(Gama_old[j]))

    # We will store updated parameters:
    mu_new = []
    Sigma_new = []
    shape_new = []
    

    for j in range(g):
        # alpha old
        mu_j, Sigma_j, shape_j, _ = alpha_prev[j]
        Delta_j_old = Delta_old[j]
        Gama_j_old = Gama_old[j]

        invGama_j_old = inv_Gama_old[j]
        invSigma_j = inv_Sigma[j]

        # 1) Compute constants: Mtij^2, Mtij
        delta_j_old = shape_j / np.sqrt(1.0 + shape_j @ shape_j)
        Mtij2 = 1.0 / (1.0 + (delta_j_old.reshape(1,-1) @ invGama_j_old @ delta_j_old.reshape(-1,1))[0,0])
        Mtij = np.sqrt(Mtij2)

        # We'll build arrays for E[i], u[i], mutij[i], A[i], etc.
        E_arr = np.zeros(n)
        u_arr = np.zeros(n)
        mutij_arr = np.zeros(n)
        dj_arr = np.zeros(n)

        
        sqrt_det_Sigma_j = np.sqrt(np.maximum(1e-300, np.linalg.det(Sigma_j)))
        
        from math import gamma as gamma_func, pi, sqrt
        from scipy.stats import t as student_t

        

        cE_num = 2.0*(nu_prev**(nu_prev/2.0))*gamma_func((p+nu_prev+1)/2.0)
        cE_den = gamma_func(nu_prev/2.0)*(pi**((p+1)/2.0))*sqrt_det_Sigma_j
        cE = cE_num / (cE_den + 1e-300)

        cU_num = 4.0*(nu_prev**(nu_prev/2.0))*gamma_func((p+nu_prev+2)/2.0)
        cU_den = gamma_func(nu_prev/2.0)* (pi**(p/2.0))*sqrt_det_Sigma_j
        cU = cU_num / (cU_den + 1e-300)

        for i in range(n):
            dif_i = X[i] - mu_j
            dj_val = dif_i @ invSigma_j @ dif_i  # scalar
            dj_arr[i] = dj_val

            mutij_val = Mtij2 * (delta_j_old @ invGama_j_old @ dif_i)
            mutij_arr[i] = mutij_val
            A_val = mutij_val / Mtij

            # pdf_j(i) = cluster_pdf[i, j]
            pdf_j_i = cluster_pdf[i, j]
            if pdf_j_i < 1e-300:
                # avoid zero
                pdf_j_i = 1e-300

            

            E_num = cE*(dj_val + nu_prev + A_val**2)**(-(p+nu_prev+1)/2.0)
            E_arr[i] = E_num / pdf_j_i  # denominator

            
            u_num = cU*(dj_val + nu_prev)**(-(p+nu_prev+2)/2.0)
            
            z_val = np.sqrt((p + nu_prev + 2)/(dj_val + nu_prev)) * A_val
            cdf_val = student_t.cdf(z_val, df=(p + nu_prev + 2))
            u_arr[i] = u_num*cdf_val / (pdf_j_i + 1e-300)

        # Now S1, S2, S3, etc.
        
        S1 = gamma[:, j]*u_arr
        
        S2 = gamma[:, j]*( mutij_arr*u_arr + Mtij*E_arr )
        
        S3 = gamma[:, j]*( mutij_arr**2*u_arr + Mtij2 + Mtij*mutij_arr*E_arr )

        # 2) Update pi_j inside _m_step . 

        # 3) M-step updates:
        
        sum_S1 = np.sum(S1)
        sum_S2 = np.sum(S2)
        sum_S1_X = np.einsum('i,ij->j', S1, X)  # weighted sum of X
        sum_S2_delta = sum_S2*Delta_j_old.ravel()  # same for Delta_j_old

        mu_j_new = (sum_S1_X - sum_S2_delta) / (sum_S1 + 1e-300)

        # Then re-define Dif with new mu_j:
        Dif = X - mu_j_new
        
        sum_S2_Dif = np.einsum('i,ij->j', S2, Dif)
        sum_S3 = np.sum(S3) 
        Delta_j_new = sum_S2_Dif / (sum_S3 + 1e-300)
        Delta_j_new = Delta_j_new.reshape(-1,1)  # make (p,1)

        
        n_j = np.sum(gamma[:, j])

        sum2 = np.zeros((p, p))
        for i in range(n):
            # Dif_i = X[i] - mu_j_new
            dif_i = Dif[i].reshape(-1,1)
            s1_val = gamma[i,j]*u_arr[i]
            s2_val = gamma[i,j]*(mutij_arr[i]*u_arr[i] + Mtij*E_arr[i])  # same as S2[i]
            s3_val = gamma[i,j]*(mutij_arr[i]**2*u_arr[i] + Mtij2 + Mtij*mutij_arr[i]*E_arr[i])  # S3[i]
            # Accumulate
            sum2 += ( s1_val*dif_i@dif_i.T
                       - s2_val*Delta_j_new@dif_i.T
                       - s2_val*dif_i@Delta_j_new.T
                       + s3_val*(Delta_j_new@Delta_j_new.T) )

        Gama_j_new = sum2 / (n_j + 1e-300)
        Sigma_j_new = Gama_j_new + Delta_j_new@Delta_j_new.T

        
        inv_sqrt_Sigma_j_new = inv(sqrtm(Sigma_j_new + 1e-10*np.eye(p)))
        numerator = inv_sqrt_Sigma_j_new @ Delta_j_new
        denom_ = 1.0 - (Delta_j_new.T @ inv(Sigma_j_new) @ Delta_j_new)
        denom_ = np.sqrt(np.maximum(1e-300, denom_[0,0]))
        shape_j_new = (numerator / denom_).ravel()  # make 1D

        mu_new.append(mu_j_new)
        Sigma_new.append(Sigma_j_new)
        shape_new.append(shape_j_new)



    def loglik_skewt_nu(nu_val):
        
        alpha_temp = []
        for j in range(g):
            alpha_temp.append((mu_new[j], Sigma_new[j], shape_new[j], nu_val))

        mix_pdf = d_mixedmvST_func(X, pi_prev, alpha_temp)
        mix_pdf = np.maximum(mix_pdf, 1e-300)
        return np.sum(np.log(mix_pdf))

    # Optimize nu in the given bounds
    
    res = minimize_scalar(
        lambda x: -loglik_skewt_nu(x),
        bounds=bounds_nu,
        method='bounded',
        options={"xatol": tol_nu}
    )
    nu_new = res.x

    # Place final alpha_new in a list of (mu, Sigma, shape, nu)
    alpha_new = []
    for j in range(g):
        alpha_new.append((mu_new[j], Sigma_new[j], shape_new[j], nu_new))

    return alpha_new




class SkewTMix(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001,initialization="kmeans",print_log_likelihood=False,max_iter=25, verbose=True):
        super().__init__(n_clusters=n_clusters, EM_type="Soft", mixture_type="identical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.initialization=initialization
        self.family = "skew_t"
        
    def _log_pdf_skewt(self,X,alphas):
        N,p=X.shape
        k=len(alphas)
        probs=np.empty((N, k))
        for j in range(k):
            alpha=alphas[j]
            with np.errstate(under="ignore"):
                probs[:, j] = np.log(dmvt_ls(X,*alpha))
        
        return probs
    
    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        return self._log_pdf_skewt(X,alpha) + np.log(pi)
    
    def fit(self,sample):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (2*self.p + (self.p*(self.p+1)/2) +1)
        np.random.seed(0)
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = k_means_init_skewt(self.data, self.k)
        elif self.initialization == "gmm":
            self.pi_not, self.alpha_not = gmm_init_skewt(self.data, self.k)
        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not
        pi_new,alpha_new, log_likelihood_new,log_gamma_new=self._fit(self.data,self.pi_temp,self.alpha_temp,estimate_alphas_skewt,pi_prev = self.pi_temp)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.cluster = log_gamma_new.argmax(axis=1)
        self.gamma_temp_ar = np.exp(log_gamma_new)
        self.log_likelihood_new = log_likelihood_new
        if self.verbose:
            print("Mixtures of Skew-t Fitting Done Successfully")
        end_time = time.time()
        self.execution_time=end_time-start_time
    
    
    
    
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster


    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)

    def get_info_mat(self):
        
        im = info_matrix(self)
        se = standard_errors_from_info_matrix(im)
        
        return im, se
        


import numpy as np
from sklearn.metrics import confusion_matrix
import copy
import matplotlib.pyplot as plt
from sklearn.metrics import f1_score, accuracy_score, precision_score, recall_score
from scipy.optimize import linear_sum_assignment

def sample_mixture_distribution(N, rand_func, pis, alphas, random_state=None):
    """
    Generate N samples from a mixture of a given probability distribution,
    along with the corresponding component labels.

    Parameters
    ----------
    N : int
        Total number of samples.
    rand_func : callable
        A function that generates random samples from a given distribution.
        It should take parameters in 'alphas[k]' and return a (M, p) array.
    pis : list or array, shape (K,)
        Mixture proportions (must sum to 1).
    alphas : list of length K
        List of parameter sets for each component, where each alphas[k]
        is a dictionary or list of parameters to be passed to `rand_func`.
    random_state : int or None, optional
        Random seed for reproducibility.

    Returns
    -------
    X : array, shape (N, p)
        Random samples from the mixture distribution.
    labels : array, shape (N,)
        Integer labels (0,1,...,K-1) indicating which component each sample belongs to.
    """
    rng = np.random.default_rng(random_state)
    K = len(pis)
    
    # Compute number of samples for each component (except last)
    sample_counts = [int(np.floor(N * pi)) for pi in pis[:-1]]
    
    # Ensure total adds up to N (assign remaining to last component)
    sample_counts.append(N - sum(sample_counts))
    
    # Generate samples from each component and store labels
    samples = []
    labels = []
    for k in range(K):
        M_k = sample_counts[k]  # Number of samples for this component
        if M_k > 0:
            samples.append(rand_func(*alphas[k], *[M_k]))  # Call the generator with params
            labels.extend([k] * M_k)  # Assign component label k to these samples
    
    # Concatenate samples and labels
    X = np.vstack(samples)
    labels = np.array(labels)

    # Shuffle samples and labels together to avoid ordering artifacts
    indices = np.arange(N)
    rng.shuffle(indices)
    X = X[indices]
    labels = labels[indices]

    return X, labels


def map_labels(true_labels, predicted_labels):
    """
    Maps predicted cluster labels to the true labels using the Hungarian algorithm
    to resolve label switching.

    Parameters:
    - true_labels: array-like, shape (n_samples,), ground-truth labels.
    - predicted_labels: array-like, shape (n_samples,), predicted labels.

    Returns:
    - mapped_predicted_labels: np.ndarray, mapped predicted labels.
    """
    true_labels, predicted_labels = np.array(true_labels), np.array(predicted_labels)
    
    unique_true_labels = np.unique(true_labels)
    unique_pred_labels = np.unique(predicted_labels)
    
    cost_matrix = np.zeros((len(unique_true_labels), len(unique_pred_labels)))

    for i, true_label in enumerate(unique_true_labels):
        for j, pred_label in enumerate(unique_pred_labels):
            match = (true_labels == true_label).astype(int)
            pred = (predicted_labels == pred_label).astype(int)
            cost_matrix[i, j] = -f1_score(match, pred)

    row_ind, col_ind = linear_sum_assignment(cost_matrix)

    label_mapping = {unique_pred_labels[col]: unique_true_labels[row] for row, col in zip(row_ind, col_ind)}
    mapped_predicted_labels = np.vectorize(label_mapping.get)(predicted_labels)

    return mapped_predicted_labels


def clustering_metrics(true_labels, predicted_labels):
    """
    Computes Accuracy, Precision, Recall, and F-score for clustering with label switching handled.

    Parameters:
    - true_labels: array-like, shape (n_samples,), ground-truth labels.
    - predicted_labels: array-like, shape (n_samples,), predicted labels.

    Returns:
    - metrics: dict containing Accuracy, Precision, Recall, and F-score.
    """
    mapped_preds = map_labels(true_labels, predicted_labels)

    metrics = {
        "accuracy": accuracy_score(true_labels, mapped_preds),
        "precision": precision_score(true_labels, mapped_preds, average='weighted', zero_division=0),
        "recall": recall_score(true_labels, mapped_preds, average='weighted', zero_division=0),
        "f_score": f1_score(true_labels, mapped_preds, average='weighted')
    }

    return metrics










def mixture_clusters(gamma_matrix, data_lol):
    max_prob_c = np.argmax(gamma_matrix, axis=1)  # Get index of maximum probability for each row
    data_cwise = [[] for _ in range(gamma_matrix.shape[1])]  # Initialize list of lists

    for i, c in enumerate(max_prob_c):
        data_cwise[c].append(data_lol[i])

    return max_prob_c, data_cwise

def plot_correlation_heatmap(covariance_matrix, variable_names):
    # Calculate correlation matrix from the covariance matrix
    correlation_matrix = np.corrcoef(covariance_matrix)

    # Create the heatmap plot
    fig, ax = plt.subplots()
    im = ax.imshow(correlation_matrix, cmap='RdYlBu', vmin=-1, vmax=1)

    # Set x-axis and y-axis tick labels
    ax.set_xticks(np.arange(len(variable_names)))
    ax.set_yticks(np.arange(len(variable_names)))
    ax.set_xticklabels(variable_names)
    ax.set_yticklabels(variable_names)

    # Rotate x-axis tick labels for better readability
    plt.setp(ax.get_xticklabels(), rotation=45, ha='right', rotation_mode='anchor')

    # Add colorbar
    cbar = ax.figure.colorbar(im, ax=ax)
    cbar.set_label('Correlation')

    # Set plot title
    ax.set_title('Correlation Heatmap')

    # Show the plot
    plt.show()
    
    
def create_cluster_dataframes(dataframe, cluster_labels):
    """
    Create different dataframes for each cluster based on the provided labels.

    Parameters:
    - dataframe: pd.DataFrame, the original dataframe containing numerical data
    - cluster_labels: list, true cluster labels for each data point

    Returns:
    - cluster_dataframes: list, a list containing dataframes for each cluster
    """

    # Check if the length of cluster_labels matches the number of rows in the dataframe
    if len(cluster_labels) != len(dataframe):
        raise ValueError("Length of cluster_labels should match the number of rows in the dataframe.")

    # Create an empty list to store dataframes for each cluster
    cluster_dataframes = []

    # Iterate through unique cluster labels
    unique_clusters = set(cluster_labels)
    for cluster in unique_clusters:
        # Filter rows belonging to the current cluster
        cluster_data = dataframe[cluster_labels == cluster]

        # Store the dataframe in the dictionary with the cluster label as the key
        cluster_dataframes.append(cluster_data)

    return cluster_dataframes


"""
Source codes are from R package MixGHD. 
Converted to python as closely as possible. Many thanks to the authors.
"""

import numpy as np
import math
from scipy.linalg import pinv, det
from scipy.special import kv  # 'kv' is the modified Bessel function of the second kind
from copy import deepcopy

# --------------------------------------------------------------------
# 1) HELPER / PLACEHOLDER FUNCTIONS (to be implemented or replaced)
# --------------------------------------------------------------------
def besselK_nuAsym(x, nu, k_max, expon_scaled=False, log=False):
    """
    Asymptotic expansion of the Bessel K_nu(x) function
    when BOTH 'nu' and 'x' are large, as described by
    Abramowitz & Stegun, eqn. 9.7.8 (p.378).

    R references:
      besselK.nuAsym(x, nu, k.max, expon.scaled=FALSE, log=FALSE)

    Parameters
    ----------
    x : float or array-like
        Argument of K_nu(x). Must be > 0 in typical usage.
    nu : float or array-like
        The order 'nu'. Must be large in typical usage.
    k_max : int
        Must be 0 <= k_max <= 5. Determines how many terms of the Debye
        polynomial expansion to use.
    expon_scaled : bool (default False)
        If True, uses an internal shift in 'eta' to emulate the scaling
        K_nu(x)*e^{+x}, as in some Bessel routines. 
    log : bool (default False)
        If True, return log(K_nu(x)) [asymptotic approximation].
        Otherwise return K_nu(x) directly.

    Returns
    -------
    out : float or ndarray
        The asymptotic approximation for K_nu(x). If 'log=True', 
        returns the approximation of log(K_nu(x)).

    Notes
    -----
    - The expansions used here assume x ~ O(nu), i.e. both large.
      Let z = x / nu. Then:
         K_nu(nu*z) ~ sqrt(pi / (2*nu)) * exp(-nu * eta) / (1 + z^2)^(1/4)
                       * {1 + d}  [with expansions up to k_max]
      where 'eta' and 'd' are computed from 'z' and 't = 1/sqrt(1+z^2)'.

    - The code is adapted directly from your R snippet. 
      For references on Debye expansions, see A&S or NIST DLMF.
    """
    # Basic checks
    if not (0 <= k_max <= 5):
        raise ValueError("k.max must be an integer in [0..5].")

    # Convert to arrays for possible broadcast
    x_arr = np.asarray(x, dtype=float)
    nu_arr = np.asarray(nu, dtype=float)
    # Determine output shape via broadcasting
    shape = np.broadcast_shapes(x_arr.shape, nu_arr.shape)
    # Broadcast to the common shape
    x_b = np.broadcast_to(x_arr, shape)
    nu_b = np.broadcast_to(nu_arr, shape)

    # Avoid division by zero if nu is zero or extremely small
    # (Though the expansion is for large nu, we still guard numeric ops.)
    nu_b_safe = np.where(nu_b == 0.0, 1e-300, nu_b)
    z = x_b / nu_b_safe  # => z = x/nu

    # sz = sqrt(1 + z^2)
    sz = np.sqrt(1.0 + z*z)
    # t = 1 / sz
    t = 1.0 / sz

    # 'eta' depends on whether we want exponent scaling
    #   If expon_scaled => eta <- 1/(z + sz), else => eta <- sz
    # Then we also add log(z / (1 + sz)) to 'eta'.
    if expon_scaled:
        # Abramowitz & Stegun eqn suggests a shift:  (sz - z) <-> 1/(z+sz)
        eta = 1.0 / (z + sz + 1e-300)
    else:
        eta = sz

    # Add the log(...) part:  eta += log(z / (1+sz))
    # We'll do that carefully to avoid log(0).
    with np.errstate(divide='ignore', invalid='ignore'):
        eta += np.log(z / (1.0 + sz) + 1e-300)

    # We'll build 'd' from expansions of the Debye polynomials:
    #   d = sum_{k=1..k_max} [ +/- u_k(t)/nu^k ]
    #   sign is slightly different from the expansions for I_nu.
    # For k_max=0, we skip expansions => d=0
    d = np.zeros(shape, dtype=float)

    if k_max > 0:
        t2 = t*t
        # u1(t) = t*(3 - 5t^2)/24
        u1_t = t*(3.0 - 5.0*t2)/24.0
        # For K_nu, the sign is negative for the first term => - u1_t / nu
        d_kmax1 = -u1_t / nu_b_safe

        if k_max == 1:
            d = d_kmax1
        else:
            # u2(t) = t^2*(81 -462t^2 + 385t^4)/1152
            u2_t = t2*(81.0 + t2*(-462.0 + 385.0*t2))/1152.0
            d_kmax2 = (d_kmax1 + (u2_t / nu_b_safe))  # all over nu => effectively => ( -u1 + u2/nu )/ nu
            d_kmax2 /= nu_b_safe

            if k_max == 2:
                d = d_kmax2
            else:
                # u3(t)
                # from R code: u3.t <- t*t2*(30375 + t2*(-369603 + t2*(765765 - 425425 t2)))/414720
                u3_t = t*t2*(30375.0 + t2*(-369603.0 + t2*(765765.0 - 425425.0*t2))) / 414720.0
                d_kmax3 = d_kmax2 + ((-u3_t) / nu_b_safe)
                d_kmax3 /= nu_b_safe

                if k_max == 3:
                    d = d_kmax3
                else:
                    # u4(t)
                    # from R code: (4465125 + t2*(...)) / 39813120
                    t4 = t2*t2
                    u4_t = t4*(4465125.0 + t2*(-94121676.0 + t2*(349922430.0 +
                            t2*(-446185740.0 + 185910725.0*t2)))) / 39813120.0
                    d_kmax4 = d_kmax3 + (u4_t / nu_b_safe)
                    d_kmax4 /= nu_b_safe

                    if k_max == 4:
                        d = d_kmax4
                    else:
                        # u5(t)
                        # from R code: t*t4*(1519035525 + t2*(-49286948607 + ...
                        u5_t = t*t4*(1519035525.0 + t2*(-49286948607.0 +
                                t2*(284499769554.0 + t2*(-614135872350.0 + 
                                t2*(566098157625.0 -188699385875.0*t2)))))  / 6688604160.0
                        d_kmax5 = d_kmax4 + ((-u5_t) / nu_b_safe)
                        d_kmax5 /= nu_b_safe

                        # k_max == 5
                        d = d_kmax5
    
    
    # We define 'res' as log or non-log
    if log:
        # log(1 + d)
        one_plus_d = 1.0 + d
        # Guard to avoid log(1 + d) negative
        with np.errstate(invalid='ignore'):
            # log1p(d)
            log1p_d = np.log1p(d)
            # If 1+d <= 0 => it might break. In practice for large z,nu, this shouldn't happen.
            mask_neg = (one_plus_d <= 0.0)
            if np.any(mask_neg):
                # fallback: for negative, set them to a large negative or handle
                log1p_d[mask_neg] = np.nan

        # -nu*eta
        neg_nu_eta = - (nu_b * eta)
        # 0.5*(log(pi/(2*nu)) - log(sz)) => note the minus sign in eqn
        # the original formula has
        #  - (log(sz) - log(pi/(2nu))) / 2
        #  = + 0.5 * ( log(pi/(2nu)) - log(sz) )
        # We'll define => halfTerm = 0.5*[log(pi/(2*nu)) - log(sz)] 
        # => Then total = log1p_d + neg_nu_eta + halfTerm
        safe_nu_b = np.where(nu_b <= 0.0, 1e-300, nu_b)
        log_term = np.log(np.pi/(2.0*safe_nu_b) + 1e-300)
        halfTerm = 0.5*(log_term - np.log(sz + 1e-300))

        res = log1p_d + neg_nu_eta + halfTerm
    else:
        # non-log
        one_plus_d = 1.0 + d
        # We do (1+d) * exp(-nu_b * eta) * sqrt(pi/(2*nu_b * sz))
        # i.e. factor = exp(-nu_b * eta) * sqrt(pi/(2*nu_b * sz))
        safe_nu_b = np.where(nu_b <= 0.0, 1e-300, nu_b)
        factor = np.sqrt(np.pi/(2.0*safe_nu_b * sz + 1e-300)) * np.exp(-nu_b*eta)
        res = one_plus_d * factor

    return res


def Rlam(x, lam=None):
    """
    Replicates the R function:
      Rlam(x, lam) = besselK(x, nu=lam+1) / besselK(x, nu=lam)
      with a fallback to besselK.nuAsym if the ratio is inf or zero.
    """
    # Ensure x and lam are NumPy arrays for consistent indexing.
    x_arr = np.atleast_1d(x)
    if lam is None:
        lam = 0.0
    lam_arr = np.atleast_1d(lam)

    # Broadcast to a common shape
    shape = np.broadcast_shapes(x_arr.shape, lam_arr.shape)
    x_b = np.broadcast_to(x_arr, shape)
    lam_b = np.broadcast_to(lam_arr, shape)

    # Evaluate K_{lam+1}(x) and K_{lam}(x)
    v1 = kv(lam_b + 1.0, x_b)
    v0 = kv(lam_b, x_b)

    # ratio
    with np.errstate(divide='ignore', invalid='ignore'):
        val = v1 / v0

    # Identify problematic elements
    set_mask = np.isinf(v1) | np.isinf(v0) | (v0 == 0) | (v1 == 0)
    if np.any(set_mask):
        # Fallback: use besselK_nuAsym or similar
        # from your_module import besselK_nuAsym  # or wherever you defined it

        x_set = x_b[set_mask]
        lam_set = lam_b[set_mask]

        lv1 = besselK_nuAsym(x=x_set, nu=(lam_set + 1.0),
                             k_max=4, expon_scaled=False, log=True)
        lv0 = besselK_nuAsym(x=x_set, nu=lam_set,
                             k_max=4, expon_scaled=False, log=True)
        val_set = np.exp(lv1 - lv0)
        val[set_mask] = val_set

    # If the original input x was a scalar, return a scalar.
    # Otherwise, return the array of values.
    if np.isscalar(x):
        return val.item()  # convert array of shape () to Python float
    return val

def logbesselKvFA(x, y):
    """
    R code:
      logbesselKvFA <- function(x, y) {
        val = log(besselK(x=y, nu=x, expon.scaled=FALSE))
        sun = is.infinite(val)
        val[sun] = besselK.nuAsym(x=y[sun], nu=abs(x[sun]), k.max=4, log=TRUE, expon.scaled=FALSE)
        return(val)
      }

    We interpret as log( K_{x}(y) ) => i.e. K_{nu=x}(y).
    If infinite, fallback to besselK.nuAsym(..., log=TRUE).
    """
    # Evaluate log( K_{x}(y) ) directly with SciPy
    # x => nu, y => argument
    # => kv(nu, y)
    # Because x,y can be arrays, let's handle them carefully.
    x_arr = np.array(x, ndmin=1)
    y_arr = np.array(y, ndmin=1)

    # We broadcast to a common shape:
    shape = np.broadcast_shapes(x_arr.shape, y_arr.shape)
    x_broad = np.broadcast_to(x_arr, shape)
    y_broad = np.broadcast_to(y_arr, shape)

    # Evaluate kv for each element
    kv_vals = kv(x_broad, y_broad)
    val = np.log(kv_vals + 1e-300)  # add small offset to avoid log(0)

    # Check infinite
    sun_mask = ~np.isfinite(val)
    if np.any(sun_mask):
        # fallback: besselK.nuAsym
        # we pass x=y_broad[sun_mask], nu=abs(x_broad[sun_mask])
        # but note your code does: besselK.nuAsym(x=y[sun], nu=abs(x[sun])...)
        # i.e. the argument is 'y' and the order is 'abs(x)'
        y_sub = y_broad[sun_mask]
        nu_sub = np.abs(x_broad[sun_mask])
        val_asym = besselK_nuAsym(x=y_sub, nu=nu_sub,
                                  k_max=4, expon_scaled=False, log=True)
        val[sun_mask] = val_asym

    return val

def besselKv(x, y):
    """
    R code:
      besselKv <- function(x, y) { besselK(x=y, nu=x) }
    i.e. just a direct call to 'besselK'.
    In Python, we do kv(x, y) = K_{x}(y).
    """
    return kv(x, y)



def numeric_grad(f, x, args=(), h=1e-8):
    """
    A simple numerical derivative (central difference)
    of f w.r.t. x, holding other args fixed.
    """
    return (f(x + h, *args) - f(x - h, *args)) / (2.0 * h)




def getall(x):
    """
    In your R code, 'getall(loglik[1:i])' is used in a loop condition, e.g.:
        while ((getall(loglik[1:i]) > 1) & (i < 100)) ...
    We guess it checks the *difference* or *relative difference* of the last two values.
    
    We'll define it as absolute difference of the last two elements.
    """
    x_clean = x[x != 0]  # remove any zeros if needed
    length = len(x_clean)
    if length < 2:
        return 9999  # big number so loop continues
    return abs(x_clean[-1] - x_clean[-2])


def combinewk(weights=None, label=None):
    """
    R code:
    combinewk <- function(weights=NULL, label=NULL){
        if (is.null(label)) stop('label is null')
        kw = label != 0
        for (j in 1:ncol(weights)) weights[kw,j] = (label == j)[kw]
        return(weights)
    }
    
    Translated to Python, noting 1-based cluster labels in R.
    """
    if label is None:
        raise ValueError("label is null")

    

    out = weights.copy()
    nrow, ncol = out.shape

    # Boolean array where label != 0
    known_mask = (label != 0)

    for j in range(1, ncol+1):
        # In R:  weights[kw, j] = (label==j)[kw]
        # Convert j to zero-based in Python -> j_idx = j-1
        j_idx = j - 1
        match_mask = (label == j)
        # wherever known_mask & match_mask is True => set to 1
        rows_to_update = np.where(known_mask & match_mask)[0]
        out[rows_to_update, :] = 0  # zero out entire row first
        out[rows_to_update, j_idx] = 1.0
    
    return out

def rgparGH(data, g=2, w=None, l=None):
    """
    R code:
    rgparGH <- function(data, g=2, w=NULL, l=NULL){
      if (is.null(w)) w = matrix(1/g, nrow=nrow(data), ncol=g)
      val = list()
      for (k in 1:g) val[[k]] = ipar(data=data, wt=w[,k], lc=l[k,])
      val$pi = rep(1/g, g)
      return(val)
    }

    In Python, we'll store them in a dict:
      val[k]   => cluster-specific parameters
      val['pi'] => mixture weights
    """
    X = np.asarray(data)
    n, p = X.shape

    if w is None:
        w = np.full((n, g), 1.0/g)

    val = {}
    
    for k in range(g):
        # l[k, ] => the k-th row of l in R => l is shape (g, p)
        lc_k = l[k, :] if l is not None else None
        val[k] = ipar(data=X, wt=w[:, k], lc=lc_k)

    val['pi'] = np.full(g, 1.0/g)
    return val

def ipar(data, wt=None, lc=None):
    """
    R version:
      ipar <- function(data, wt=NULL, lc=NULL){
         if (is.null(wt)) wt = rep(1,nrow(data))
         p = ncol(data)
         val = list()
         val$mu    = lc
         val$alpha = rep(0,p)
         val$sigma = diag( diag( cov.wt(data, wt=wt)$cov ) )
         ... ensure diag>0.1 ...
         val$cpl   = c(1, -1/2)
         return(val)
      }
    """
    X = np.asarray(data)
    n, p = X.shape
    
    if wt is None:
        wt = np.ones(n)
    
    # We'll build a dictionary for the return value
    val = {}
    val['mu'] = np.array(lc)  # shape (p,), as in R
    val['alpha'] = np.zeros(p)
    
    
    w_sum = np.sum(wt)
    if w_sum < 1e-12:
        w_sum = 1e-12
    
    mean_w = np.einsum('ij,i->j', X, wt) / w_sum
    X_centered = X - mean_w
    cov_w = np.einsum('i,ij,ik->jk', wt, X_centered, X_centered) / w_sum
    
    diag_cov = np.diag(np.diag(cov_w))
    
    # "if (p==1) val$sigma=var(data)"
    if p == 1:
        diag_cov = np.array([[np.var(X)]])
    else:
        # The R code ensures positivity on diagonal, at least 0.1
        for i in range(p):
            if diag_cov[i, i] < 0.1:
                diag_cov[i, i] = 0.1
    
    # "if any eigen(val$sigma)$values <= 0 ) val$sigma= diag(apply(data,2,var))"
    # We'll do a quick check:
    eigvals = np.linalg.eigvals(diag_cov)
    if np.any(eigvals <= 0):
        # fallback
        diag_cov = np.diag(np.var(X, axis=0))
    
    val['sigma'] = diag_cov
    
    # cpl = c(1, -1/2)
    val['cpl'] = [1.0, -0.5]
    return val

def igpar(data=None, g=None, method="kmeans", nr=None, label=None):
    """
    R code:
    igpar <- function(data=NULL, g=NULL, method="kmeans", nr=NULL, label=NULL){
        gpar = igpar3(data=data, g=g, n=10, method=method, nr=nr, label=label)
        return(gpar)
    }
    We'll define igpar3 as a placeholder.
    """
    return igpar3(data=data, g=g, n=10, method=method, nr=nr, label=label)

def igpar3(data=None, g=None, n=10, label=None, method="kmeans", nr=10):
    """
    R version (truncated explanation):
      - If g==1, do a "kmeans(data, 1)" or just get the global mean => then run a small EM
      - Else handle different 'method' values:
          "modelBased" => calls gpcm(...)  [ placeholder ]
          "hierarchical" => calls hclust => cutree => ...
          "random" => random initialization repeated nr times
          "kmedoids" => pam(...)
          else => kmeans(...) 
        then run a small EM loop

    We'll add placeholders for gpcm, pam, hclust, etc. Can be implemented later
    """
    X = np.asarray(data)
    nrows, p = X.shape

    # Helper placeholders for cluster code:
    def hclust_cutree(X, g):
        """
        Minimal placeholder for hierarchical clustering with Ward's method.
        In practice, you'd use e.g. scipy's linkage(..., method='ward') & fcluster(...).
        We just do a random label assignment as a dummy fallback.
        """
        # from scipy.cluster.hierarchy import linkage, fcluster
        # Z = linkage(X, method='ward')
        # clusters = fcluster(Z, t=g, criterion='maxclust')
        # return clusters  # 1-based cluster labels
        labs = np.random.randint(1, g+1, size=len(X))
        return labs
    
    def kmeans_init(X, g):
        """
        Minimal placeholder for k-means -> return (centers, labels).
        You can replace with scikit-learn's KMeans or similar.
        """
        from sklearn.cluster import KMeans
        km = KMeans(n_clusters=g, n_init=1, random_state=42).fit(X)
        return (km.cluster_centers_, km.labels_+1)  # scikit-learn gives 0-based => add 1
        # return (km.cluster_centers_, km.labels_)  # scikit-learn gives 0-based => add 1
    
    def pam(X, g):
        """
        Minimal placeholder for 'pam'. In R, pam() returns medoids, clustering, ...
        We'll just reuse k-means centers as a rough stand-in.
        """
        centers, labels = kmeans_init(X, g)
        # We'll pretend centers are "medoids."
        return {
            'medoids': centers,
            'clustering': labels
        }
    
    def gpcm(X, G, mnames):
        """
        Placeholder for 'gpcm(data, G=g, mnames=c("VVV"))' from 'modelBased'.
        We'll just mimic a result with random or KMeans approach.
        We assume it returns something with $map and $gpar[[k]]$mu
        """
        # Just do a k-means approach for demonstration
        centers, labels = kmeans_init(X, G)
        # Build a fake object
        out = {}
        out['map'] = labels  # 1-based
        out['gpar'] = []
        for k in range(G):
            out['gpar'].append({'mu': centers[k]})
        return out

    # Now replicate the logic from R:
    if g == 1:
        # "lk = kmeans(data, 1)" => effectively the global mean
        # We'll do the same with "lc = apply(data, 2, mean)"
        lc = np.mean(X, axis=0).reshape(1, p)
        l = np.ones(nrows, dtype=int)  # all in cluster 1
        # combinewk
        # from . import combinewk, rgparGH, EMgrstepGH  # or adapt import
        z = combinewk(weights=np.zeros((nrows, g)), label=l)
        gpar = rgparGH(data=X, g=g, w=z, l=lc)
        # small EM loop
        for j in range(n):
            try:
                gpar = EMgrstepGH(data=X, gpar=gpar, v=1, label=l, w=z)
            except:
                pass
        return gpar
    else:
        # method-based logic
        if method == "modelBased":
            # call gpcm
            lk = gpcm(X, G=g, mnames=["VVV"])
            l = lk['map']
            lc = lk['gpar'][0]['mu'].reshape(1, p)
            for i in range(2, g+1):
                mu_i = lk['gpar'][i-1]['mu'].reshape(1, p)
                lc = np.vstack([lc, mu_i])
            
            # from . import combinewk, rgparGH, EMgrstepGH
            z = combinewk(weights=np.zeros((nrows, g)), label=l)
            gpar = rgparGH(data=X, g=g, w=z, l=lc)
            
            for j in range(n):
                try:
                    gpar = EMgrstepGH(data=X, gpar=gpar, v=1, label=l, w=z)
                except:
                    pass
            return gpar
        
        elif method == "hierarchical":
            l = hclust_cutree(X, g)
            # compute means
            lc_list = []
            for cluster_id in range(1, g+1):
                cluster_pts = X[l==cluster_id]
                if len(cluster_pts) > 0:
                    lc_list.append(cluster_pts.mean(axis=0))
                else:
                    # fallback
                    lc_list.append(np.mean(X, axis=0))
            lc = np.vstack(lc_list)
            
            # from . import combinewk, rgparGH, EMgrstepGH
            z = combinewk(weights=np.zeros((nrows, g)), label=l)
            gpar = rgparGH(data=X, g=g, w=z, l=lc)
            for j in range(n):
                try:
                    gpar = EMgrstepGH(data=X, gpar=gpar, v=1, label=l, w=z)
                except:
                    pass
            return gpar
        
        elif method == "random":
            # repeated random assignment
            # from . import combinewk, rgparGH, EMgrstepGH, llikGH
            llkO = -np.inf
            gpar_best = None
            
            for r_i in range(nr):
                # random label in 1..g
                l_rand = np.random.randint(1, g+1, size=nrows)
                # compute means
                lc_list = []
                for cluster_id in range(1, g+1):
                    cluster_pts = X[l_rand==cluster_id]
                    if len(cluster_pts) > 0:
                        lc_list.append(cluster_pts.mean(axis=0))
                    else:
                        lc_list.append(np.mean(X, axis=0))
                lc = np.vstack(lc_list)
                
                z = combinewk(weights=np.zeros((nrows, g)), label=l_rand)
                gparO = rgparGH(data=X, g=g, w=z, l=lc)
                
                loglik = np.zeros(100)
                i = 0
                for i in range(3):
                    try:
                        gparO = EMgrstepGH(data=X, gpar=gparO, v=1, label=l_rand, w=z)
                    except:
                        pass
                    loglik[i] = llikGH(X, gparO)
                
                while (getall(loglik[:i+1]) > 1) and (i < 99):
                    i += 1
                    try:
                        gparO = EMgrstepGH(data=X, gpar=gparO, v=1, label=l_rand, w=z)
                    except:
                        pass
                    loglik[i] = llikGH(X, gparO)
                
                llk = llikGH(X, gparO)
                if llk > llkO:
                    llkO = llk
                    gpar_best = deepcopy(gparO)
            
            return gpar_best
        
        elif method == "kmedoids":
            # call pam
            # from . import pam, combinewk, rgparGH, EMgrstepGH
            lk = pam(X, g)
            lc = lk['medoids']
            l = lk['clustering']
            
            z = combinewk(weights=np.zeros((nrows, g)), label=l)
            gpar = rgparGH(data=X, g=g, w=z, l=lc)
            
            for j in range(n):
                try:
                    gpar = EMgrstepGH(data=X, gpar=gpar, v=1, label=l, w=z)
                except:
                    pass
            return gpar
        
        else:
            # default to "kmeans"
            # from . import combinewk, rgparGH, EMgrstepGH
            centers, l = kmeans_init(X, g)
            # centers => shape (g, p), l => shape (n,) w/ 1..g
            z = combinewk(weights=np.zeros((nrows, g)), label=l)
            gpar = rgparGH(data=X, g=g, w=z, l=centers)
            
            for j in range(n):
                try:
                    gpar = EMgrstepGH(data=X, gpar=gpar, v=1, label=l, w=z)
                except:
                    pass
            return gpar


def weightsGH(data=None, gpar=None, v=1):
    """
    R code:
      weightsGH <- function(data=NULL, gpar=NULL, v=1){
        G = length(gpar$pi)
        if (G>1){
          zlog = matrix(0, nrow=nrow(data), ncol=length(gpar$pi))
          for (k in 1:G) zlog[,k] = ddghypGH(..., log=TRUE)
          w = t(apply(zlog, 1, function(z,wt,v){
             fstar = v*(z + log(wt)) - max(v*(z + log(wt)))
             x = exp(fstar)
             if(sum(x)==0) x=rep(1,length(x))
             x = x/sum(x)
             return(x)
          }, wt=gpar$pi, v=v ))
        } else w=matrix(1,nrow(data), ncol=G)
        return(w)
      }

    We'll replicate that in Python, using our `ddghypGH` function for the GH log density.
    """
    X = np.asarray(data)
    n, p = X.shape
    G = len(gpar['pi'])

    if G > 1:
        zlog = np.zeros((n, G))
        for k in range(G):
            # ddghypGH with log=True
            lval = ddghypGH(X, gpar[k], log=True, invS=None)
            zlog[:, k] = lval
        
        # Now replicate 'apply(zlog, 1, function(z, wt, v){...})'
        w_rows = []
        for i in range(n):
            z = zlog[i, :]
            wt = gpar['pi']
            # fstar = v*(z + log(wt)) - max( v*(z + log(wt)) )
            z_plus_logwt = z + np.log(wt + 1e-300)
            temp = v * z_plus_logwt
            fstar = temp - np.max(temp)
            x = np.exp(fstar)
            if np.sum(x) == 0:
                x = np.ones_like(x)
            x /= np.sum(x)
            w_rows.append(x)
        w = np.vstack(w_rows)
    else:
        w = np.ones((n, G))
    return w


def MAPGH(data, gpar, label=None):
    """
    R code:
    MAPGH <- function(data, gpar, label=NULL){
      w = weightsGH(...)
      if(!is.null(label)) w = combinewk(weights=w, label=label)
      z = apply(w, 1, function(z){ z=(1:length(z))[z==max(z)]; return(z[1])})
      z = as.numeric(z)
      return(z)
    }
    """
    w = weightsGH(data=data, gpar=gpar, v=1)
    if label is not None:
        w = combinewk(weights=w, label=label)
    
    # For each row i, find the cluster k with the maximum w[i,k].
    # In R we do 1-based, but let's keep it consistent:
    # z = which max => 1-based
    n = w.shape[0]
    z = np.zeros(n, dtype=int)
    for i in range(n):
        row = w[i, :]
        kmax = np.argmax(row)  # 0-based index
        # z[i] = kmax + 1        # store 1-based
        z[i] = kmax
    return z



# --------------------------------------------------------------------
# 2) UTILITY FUNCTIONS directly translated from your R code
# --------------------------------------------------------------------

def weighted_sum(z, wt):
    """R’s weighted.sum(z, wt=...) -> sum(z*wt)."""
    return np.sum(z * wt)


def ddghypGH(x, par, log=False, invS=None):
    """
    R function ddghypGH(...) for the generalized hyperbolic distribution's density.

    Parameters
    ----------
    x   : (n, p) data array
    par : dictionary with keys ['mu', 'sigma', 'alpha', 'cpl']
          cpl[0] ~ 'omega', cpl[1] ~ 'lambda'
    log : bool, if True return log of pdf
    invS: optional inverse of sigma
    """
    # Convert to np.array if needed
    X = np.asarray(x)
    if X.ndim == 1:
        X = X.reshape(1, -1)
    n, p = X.shape
    
    mu = par['mu']
    sigma = par['sigma']
    alpha = par['alpha']
    
    # ensure shapes
    mu = np.asarray(mu).ravel()        # shape (p,)
    alpha = np.asarray(alpha).ravel()  # shape (p,)
    
    omega_value = max(par.get('cpl', [1e-10])[0], 1e-10)  # Ensure a positive value
    omega = math.exp(math.log(omega_value))

    # omega = math.exp(math.log(par['cpl'][0]))  # = par$cpl[1] in R, but they did exp(log(...))
    lam   = par['cpl'][1]  # lambda
    
    if invS is None:
        invS = pinv(sigma)  # pseudo-inverse if needed
    
    # pa = omega + alpha' invS alpha
    alpha_invSa = alpha @ invS @ alpha
    
    # Compute distances
    # mahalanobis(x, center=mu, invS)
    # We'll do it manually:
    X_centered = X - mu
    mx = np.einsum('ij,jk,ik->i', X_centered, invS, X_centered)
    # mx is length n, each row's Mahalanobis distance
    
    pa = omega + alpha_invSa
    pa2 = np.full(n, pa)
    kx = np.sqrt(mx * pa2)
    
    # xmu%*%invS%*%alpha => same approach:
    xmua = np.einsum('ij,jk->ik', X_centered, invS)  # shape (n,p)
    lvx3 = np.einsum('ij,j->i', xmua, alpha)         # shape (n,)
    
    # Now piece together the log of the GH density
    # R code built an array lvx with columns [ (lambda - d/2)*log(kx), log(besselK(...)) - kx, xmu%*%..., 0? ]
    # Then it sums them plus some constants in lv.
    
    lvx = np.zeros((n, 4))
    
    lvx[:, 0] = (lam - p/2.0) * np.log(kx + 1e-300)
    # besselK(kx, nu=lam - d/2) => in python: kv(lam - p/2, kx)
    # The R code used `expon.scaled=TRUE` and then subtracted kx. We replicate that:
    # => log(besselK(..., expon.scaled=TRUE)) - kx
    # We skip "expon.scaled=TRUE" because python's kv does not have that mode directly.
    
    # We can do:
    #   logBessel = np.log(kv(lam - p/2, kx) + 1e-300)
    #   lvx[:,1] = logBessel
    # but the R code subtracts kx:
    logBessel = np.log(kv(lam - p/2.0, kx) + 1e-300)
    lvx[:, 1] = logBessel - kx
    
    lvx[:, 2] = lvx3  # xmu %*% invS %*% alpha
    lvx[:, 3] = 0.0   # the code put 0 for that column, but effectively we might not need it
    
    # Summation of these is lvx_row_sum
    lvx_row_sum = np.sum(lvx, axis=1)
    
    # The 'lv' array in R has length 6 with various constants. We'll collect them similarly:
    lv = np.zeros(6)
    
    # 1) -1/2 * log(det(sigma)) - d/2 * (log(2) + log(pi))
    # watch out for det(sigma) possibly zero. Use a fallback or pinv if needed
    # or we just do math.log(det) if non-singular
    try:
        dsig = det(sigma)
        if dsig <= 0:
            dsig = 1e-300
    except:
        dsig = 1e-300
    
    lv[0] = -0.5 * math.log(dsig) - (p/2.0)*(math.log(2.0) + math.log(math.pi))
    
    # 2)  omega - log(besselK(omega, nu=lambda, expon.scaled=TRUE))
    #    again no direct 'expon.scaled' in Python
    logK_omega = np.log(kv(lam, omega) + 1e-300)
    lv[1] = omega - logK_omega
    
    # 3) -lambda/2*( log(1) )
    #   that is zero, because log(1)=0 => so no effect
    lv[2] = 0.0
    
    # 4) lambda*log(omega)*0 => also zero
    lv[3] = 0.0
    
    # 5) (d/2 - lambda)*log(pa)
    lv[4] = (p/2.0 - lam)*math.log(pa + 1e-300)
    
    # Summation of those 6 constants is sum_lv
    sum_lv = np.sum(lv)
    
    val = lvx_row_sum + sum_lv  # shape (n,)
    if not log:
        val = np.exp(val)
    return val

def gig2GH(x, par, invS=None):
    """
    R function gig2GH(). 
    Returns a matrix of shape (n, 3) with [w, invw, logw].
    Each row i is E[W], E[1/W], E[logW] for observation x_i under GIG.
    """
    X = np.asarray(x)
    if X.ndim == 1:
        X = X.reshape(1, -1)
    n, d = X.shape
    
    if invS is None:
        invS = pinv(par['sigma'])
    
    alpha = par['alpha']
    alpha = np.asarray(alpha).ravel()
    
    # par$cpl[1] = "omega" in exponent form
    # omega = math.exp(math.log(par['cpl'][0]))
    omega_value = max(par.get('cpl', [1e-10])[0], 1e-10)  # Ensure a positive value
    omega = math.exp(math.log(omega_value))
    # lambda = par$cpl[2]
    lam = par['cpl'][1]
    
    # a1 = omega + alpha' invS alpha
    alpha_invSa = alpha @ invS @ alpha
    a1 = omega + alpha_invSa
    
    # b1 = omega + mahalanobis(x, mu, invS)
    # we interpret 'mahalanobis(x, center=par$mu, invS=..., inverted=TRUE)'
    mu = np.asarray(par['mu']).ravel()
    Xc = X - mu
    b1 = np.einsum('ij,jk,ik->i', Xc, invS, Xc) + omega
    
    # v1 = par$cpl[2] - d/2
    v1 = lam - d/2.0
    
    # Now call gigGH(a=a1, b=b1, v=v1)
    val = gigGH(a=a1, b=b1, v=v1)
    return val

def gigGH(a=None, b=None, v=None):
    """
    R function gigGH(). 
    Returns a matrix (n, 3) with columns [w, invw, logw].
    Interprets the formula:
       w    = K_{v+1}( sqrt(a*b) ) / K_v( sqrt(a*b) ) * sqrt(b/a)   ... etc.
       invw = ...
       logw = ...
    """
    # a, b can be scalars or arrays. We’ll treat them as arrays.
    # We want elementwise operations
    if not isinstance(a, np.ndarray):
        a = np.array([a])
    if not isinstance(b, np.ndarray):
        b = np.array([b])
    
    # Ensure same shape
    # The R code expects 'length(a)' to match 'length(b)' => the number of observations
    n = max(a.size, b.size)
    A = np.full(n, a.item()) if a.size==1 else a.ravel()
    B = np.full(n, b.item()) if b.size==1 else b.ravel()
    
    sab = np.sqrt(A*B)  # shape (n,)
    # kv( nu=v+1, x=sab ) / kv( nu=v, x=sab )
    # For numeric stability, add small offset if needed.
    eps = 1e-300
    
    # We replicate R's use of 'besselK(..., expon.scaled=TRUE)' by just normal kv
    # If sab is large, might need log or scaling. For now, do direct kv:
    kv_num = kv(v+1.0, sab+eps)
    kv_den = kv(v,     sab+eps) + eps
    kv_ratio = kv_num / kv_den
    
    sb_a = np.sqrt(B/(A+eps))  # shape (n,)
    
    w    = kv_ratio * sb_a
    invw = kv_ratio*(1.0/sb_a) - 2.0*v/B  # from your R code
    # For logw, the code uses a numeric derivative approach: grad(logbesselKv, x=rep(v,...), y=sab,...) 
    # Currently your code calls 'grad(logbesselKv, ...)'. We do not replicate that here exactly.
    # We'll do a placeholder:
    logw = np.log(sb_a+eps)  # plus a derivative term that we have not fully replicated
    
    # Put together as (n,3)
    out = np.column_stack((w, invw, logw))
    return out

def llikGH(data, gpar, delta=0):
    """
    R function llikGH(). 
    Computes the log-likelihood for mixture of GH distributions in gpar.

    gpar is structured:
      gpar['pi'] : array of shape (G,)
      gpar[k]    : dict with keys ['mu','sigma','alpha','cpl'] for cluster k
    """
    n, p = data.shape
    G = len(gpar['pi'])
    
    logz = np.zeros((n, G))
    for k in range(G):
        logz[:, k] = ddghypGH(data, gpar[k], log=True, invS=None)
    
    
    total = 0.0
    for i in range(n):
        # sum(exp(z)*wt) => sum(exp(logz[i,:])* gpar$pi )
        ex = np.exp(logz[i, :])
        weighted = ex * gpar['pi']
        total += math.log(np.sum(weighted) + 1e-300)
    return total


# --------------------------------------------------------------------
# 3) The CORE EM functions: mainMGHD & EMgrstepGH & updatemaScpl
# --------------------------------------------------------------------

def mainMGHD(data=None, gpar0=None, G=None, n=None, label=None, eps=1e-5, method=None, nr=None):
    """
    R function mainMGHD().
    data  : (N, p) array
    gpar0 : initial gpar (dictionary) or None
    G     : number of clusters
    n     : maximum number of iterations? (R uses n for the loop limit)
    label : optional cluster labels
    eps   : tolerance
    method: method name
    nr    : not used except possibly for 'igpar'
    """
    # pcol = ncol(data)
    N, pcol = data.shape
    
    # If label is given, compute cluster means for initialization
    if label is not None:
        # in R code: we do an apply(data[label==1,],2,mean), etc.
        # We'll store those in lc
        unique_labels = np.unique(label)
        # Make sure labeling is consistent with G
        # The R code just picks the means for each i in 1..G
        lc = []
        for i in range(1, G+1):
            cluster_pts = data[label == i]
            if cluster_pts.shape[0] > 0:
                mean_i = cluster_pts.mean(axis=0)
            else:
                mean_i = np.zeros(pcol)
            lc.append(mean_i)
        lc = np.vstack(lc)
        
        # combinewk
        z = combinewk(weights=np.full((N, G), 1.0/G), label=label)
        # If gpar0 not given, then call rgparGH
        if gpar0 is None:
            gpar = rgparGH(data=data, g=G, w=z, l=lc)
        else:
            gpar = gpar0
    else:
        # no label
        if gpar0 is None:
            # call igpar
            gpar = igpar(data=data, g=G, method=method, nr=nr)
        else:
            gpar = gpar0
    
    loglik = np.zeros(n, dtype=float)
    
    # initial steps
    i = 0
    for i in range(3):
        gpar = EMgrstepGH(data=data, gpar=gpar, v=1, label=label, w=None)
        loglik[i] = llikGH(data, gpar)
    
    # main loop
    # while( ( getall(loglik[1:i]) > eps) & (i < n ) )
    # "getall" presumably checks difference in consecutive loglik or something. Let's replicate that logic:
    def rel_change(x):
        if len(x) < 2:
            return 999.0
        return abs(x[-1] - x[-2]) / (abs(x[-2]) + 1e-300)
    
    i = 2  # after the first 3 steps, i is at index 2
    while True:
        if rel_change(loglik[:i+1]) < eps:
            break
        if (i+1) >= (n-1):
            break
        
        i += 1
        gpar = EMgrstepGH(data=data, gpar=gpar, v=1, label=label, w=None)
        loglik[i] = llikGH(data, gpar)
    
    # if i<n => shorten loglik
    if i < (n-1):
        loglik = loglik[:i+1]
    
    # compute BIC, ICL, AIC, AIC3
    # BIC=2*loglik[i]-log(nrow(data))*((G-1)+G*(2*pcol+2+pcol*(pcol-1)/2))
    # We replicate that:
    ndata = data.shape[0]
    # number of parameters for GH mixture:
    #   (G-1) + G*(2*pcol+2 + pcol*(pcol-1)/2)
    #   = (G-1) + G*(2pcol + 2 + pcol(pcol-1)/2)
    nparams = (G - 1) + G*(2*pcol + 2 + pcol*(pcol-1)/2.0)
    
    ll_final = loglik[i]
    BIC = 2*ll_final - math.log(ndata)*nparams
    # z = weightsGH(...)
    z = weightsGH(data=data, gpar=gpar, v=1)
    # ICL = BIC + 2*sum(log(apply(z,1,max)))
    row_max = np.max(z, axis=1)
    ICL = BIC + 2.0*np.sum(np.log(row_max + 1e-300))
    
    # AIC=2*ll_final - 2*nparams
    AIC  = 2*ll_final - 2.0*nparams
    AIC3 = 2*ll_final - 3.0*nparams
    
    # map = MAPGH(...)
    map_labels = MAPGH(data=data, gpar=gpar, label=label)
    
    val = {
        'loglik': loglik,
        'gpar': gpar,
        'z': z,
        'map': map_labels,
        'BIC': BIC,
        'ICL': ICL,
        'AIC': AIC,
        'AIC3': AIC3
    }
    return val

def EMgrstepGH(data=None, gpar=None, v=1, label=None, w=None):
    """
    R function EMgrstepGH(...).
    - E-step: compute w=weightsGH(...) unless it's given
    - If label given, combinewk
    - M-step: for each cluster k, call updatemaScpl
    - Then update gpar$pi
    """
    if w is None:
        w = weightsGH(data=data, gpar=gpar, v=v)
    if label is not None:
        w = combinewk(weights=w, label=label)
    
    G = len(gpar['pi'])
    d = len(gpar[0]['mu'])  # dimension from the first component
    
    # For each cluster k in 1..G:
    for k in range(G):
        invS = None
        alpha_known = None
        gpar[k] = updatemaScpl(x=data, par=gpar[k], weights=w[:, k], invS=invS, alpha_known=alpha_known, v=v)
    
    # update gpar$pi = colMeans of w
    gpar['pi'] = np.mean(w, axis=0)
    return gpar

def updatemaScpl(x, par, weights=None, invS=None, alpha_known=None, v=None):
    """
    R function updatemaScpl(...).
    - Compute 'abc = gig2GH(...)'
    - Then update mu, alpha, sigma, cpl
    """
    X = np.asarray(x)
    n, p = X.shape
    if weights is None:
        weights = np.ones(n)
    if invS is None:
        invS = pinv(par['sigma'])
    
    abc = gig2GH(x=X, par=par, invS=invS)  # shape (n, 3) => columns [w, invw, logw]
    # Weighted average of each column:
    sumw = np.sum(weights)
    
    # ABC = apply(abc, 2, weighted.sum, wt=weights)/sumw => array( [A, B, C], shape=(3,) )
    ABC = np.array([np.sum(abc[:, j]*weights) for j in range(3)]) / sumw
    A = ABC[0]
    B = ABC[1]
    
    
    
    w_col = abc[:, 0]
    invw_col = abc[:, 1]
    # logw_col = abc[:, 2]
    
    # u = (B - abc[,2])*weights => in R, "abc[,2]" was 'invw' => so (B - invw[i])
    u = (B - invw_col) * weights
    # t = (A*abc[,2] - 1)*weights => (A*invw_col - 1)*weights
    t_ = (A*invw_col - 1.0) * weights
    T_ = np.sum(t_)
    
    if alpha_known is None:
        # mu.new
        # Weighted sum of X columns with weights t_
        mu_num = np.einsum('ij,i->j', X, t_)
        if abs(T_) < 1e-12:
            mu_new = par['mu']
        else:
            mu_new = mu_num / T_
        
        # alpha.new
        alpha_num = np.einsum('ij,i->j', X, u)
        if abs(T_) < 1e-12:
            alpha_new = par['alpha']
        else:
            alpha_new = alpha_num / T_
    else:
        alpha_new = alpha_known
        # mu.new = weighted.mean(X, w=abc[,2]*weights) - alpha.new/ABC[2]
        
        invw_weights = invw_col * weights
        sum_invw_weights = np.sum(invw_weights)
        mu_mean = np.einsum('ij,i->j', X, invw_weights) / (sum_invw_weights + 1e-12)
        mu_new = mu_mean - alpha_new / B   # since B=ABC[2] in R code? Actually B=ABC[2], but R code = invw col. Let's keep it consistent.
    
    # alpha.new = alpha.new * v => from the R code: alpha.new = alpha.new * v
    alpha_new = alpha_new * v
    
    
    w_for_cov = invw_col * weights
    sum_w_for_cov = np.sum(w_for_cov)
    
    X_centered = X - mu_new
    # Weighted cov
    if sum_w_for_cov > 1e-12:
        covML = np.einsum('i,ij,ik->jk', w_for_cov, X_centered, X_centered) / sum_w_for_cov
    else:
        covML = np.eye(p)*1e-5
    A_ = covML * B  # "A" in R code
    
    
    R = A_
    
    for i_col in range(p):
        if R[i_col, i_col] < 1e-5:
            R[i_col, i_col] = 1e-5
    
    # par.old = c( log(par$cpl[1]) , par$cpl[2] )
    # We do the same:
    cpl1 = par['cpl'][0]  # in R, cpl[1], then we do log(cpl[1])
    cpl2 = par['cpl'][1]  # in R, cpl[2]
    # try:
    #     cpl1 = np.maximum(cpl1[0],1)
    # except:
    #     cpl1 = np.maximum(cpl1,1)
    # print(cpl1)
    par_old = [math.log1p(cpl1), cpl2] #log giving error when it becomes zero!
    
    
    def updateol(ol=None, ABC=None, n=1):
  
    
        if ol is None or ABC is None:
            return ol
    
        ol = list(ol)  # convert to mutable list if needed
    
        for i in range(n):
            # If ABC[2] == 0 => ol[1] = 0
            if abs(ABC[2]) < 1e-15:
                ol[1] = 0.0
            else:
                
    
                # We'll define a local function f(lam_value) => logbesselKvFA(lam_value, omega=ol[0])
                def f_for_grad(lam_value, omega_val):
                    return logbesselKvFA(lam_value, omega_val)
    
                # numeric derivative wrt lam_value
                bv = numeric_grad(f_for_grad, ol[1], args=(ol[0],))
    
                # Then: ol[2] = ABC[3]*(ol[2]/bv)
                # i.e. "lambda_new = ABC[3] * (lambda_old / bv)"
                # in R: ol[2] => lam, ABC[3] => ABC[2] in 0-based
                lam_old = ol[1]
                if abs(bv) < 1e-15:
                    # avoid division by zero
                    pass
                else:
                    ol[1] = ABC[2] * (lam_old / bv)
    
            # R code calls:
            #   Rp = Rlam(ol[1], lam=+ol[2])
            #   Rn = Rlam(ol[1], lam=-ol[2])
            # But remember in R: ol[1] is the first element => we interpret as "omega"?
            # Actually the snippet suggests "ol[1]" is "omega" and "ol[2]" is "lambda".
            # Let's keep that:
            omega_now = ol[0]
            lam_now   = ol[1]
    
            Rp = Rlam(omega_now, lam=+lam_now)
            Rn = Rlam(omega_now, lam=-lam_now)
    
            # f1 = Rp + Rn - (ABC[1] + ABC[2]) => in R => ABC[1], ABC[2]
            f1 = Rp + Rn - (ABC[0] + ABC[1])
    
            
            f2_part1 = Rp**2 - ((2*lam_now + 1)/ (omega_now+1e-15)) * Rp - 1
            f2_part2 = Rn**2 - ((2*(-1*lam_now) + 1)/ (omega_now+1e-15)) * Rn - 1
            f2_val = f2_part1 + f2_part2
    
            
            if (ol[0] - f1/(f2_val+1e-15)) > 0:
                ol[0] = ol[0] - f1/(f2_val+1e-15)
    
        return ol

    
    
    a_ = updateol(par_old, ABC=ABC, n=2)
    cpl_new = [a_[0], a_[1]]
    
    new_par = {
        'mu': mu_new,
        'alpha': alpha_new,
        'sigma': R,
        'cpl': cpl_new
    }
    return new_par



import numpy as np
# import cupy as np
import pandas as pd
from scipy.special import logsumexp
from abc import ABCMeta, abstractmethod
from fmvmm.utils import utils_mixture
# import numpy

class BaseMixture(metaclass=ABCMeta):

    def __init__(self, n_clusters, EM_type, mixture_type, tol, print_log_likelihood, max_iter, verbose):
        self.n_clusters = n_clusters
        self.EM_type = EM_type
        self.mixture_type = mixture_type
        self.tol = tol
        self.print_log_likelihood = print_log_likelihood
        self.max_iter = max_iter
        self.verbose = verbose
        self.fitted=False
        
    def _process_data(self,X):
        if isinstance(X, pd.DataFrame):
            return X.values
        elif isinstance(X, np.ndarray) and len(X.shape) == 2:
            return X
        elif isinstance(X, list):
            if all(isinstance(row, list) for row in X):
                return np.array(X)
            else:
                raise ValueError("Input should be a list of lists for conversion to 2D NumPy array.")
        else:
            raise ValueError("Input should be a Pandas DataFrame, a 2D NumPy array, or a list of lists.")

    # @abstractmethod
    # def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
    #     pass

    # @abstractmethod
    # def _estimate_weighted_log_prob_nonidentical(self, X, alpha, pi, dist_comb):
    #     pass

    def _estimate_log_prob_resp_identical(self, X, alpha, pi):
        weighted_log_prob = self._estimate_weighted_log_prob_identical(
            X, alpha, pi)
        log_prob_norm = logsumexp(weighted_log_prob, axis=1)
        with np.errstate(under="ignore"):
            # ignore underflow
            log_resp = weighted_log_prob - log_prob_norm[:, np.newaxis]
        return log_prob_norm, log_resp

    def _estimate_log_prob_resp_nonidentical(self, X, alpha, pi, dist_comb):
        weighted_log_prob = self._estimate_weighted_log_prob_nonidentical(
            X, alpha, pi, dist_comb)
        log_prob_norm = logsumexp(weighted_log_prob, axis=1)
        with np.errstate(under="ignore"):
            # ignore underflow
            log_resp = weighted_log_prob - log_prob_norm[:, np.newaxis]
        return log_prob_norm, log_resp

    def _e_step(self, X, alpha, pi, dist_comb=None):
        if self.mixture_type == "identical":
            log_prob_norm, log_resp = self._estimate_log_prob_resp_identical(
                X, alpha, pi)
        elif self.mixture_type == "nonidentical":
            log_prob_norm, log_resp = self._estimate_log_prob_resp_nonidentical(
                X, alpha, pi, dist_comb)
        return np.sum(log_prob_norm), log_resp
    
    
    def _fit(self,X,pi_temp,alpha_temp,estimate_alphas_function,dist_comb=None,**kwargs):
        X=self._process_data(X)
        self.p = X.shape[1]
        self.n = X.shape[0]
        self.log_likelihoods=[]
        stop=False
        while stop!=True:
            log_likelihood_old, log_gamma_temp = self._e_step(
                X, alpha_temp, pi_temp,dist_comb)
            if len(self.log_likelihoods)==0:
                self.log_likelihoods.append(log_likelihood_old)
            gamma_temp_ar=np.exp(log_gamma_temp)
            pi_new,alpha_new=self._m_step(gamma_temp_ar, X, estimate_alphas_function,dist_comb, **kwargs)
            log_likelihood_new, log_gamma_new = self._e_step(
                X, alpha_new, pi_new,dist_comb)
            self.log_likelihoods.append(log_likelihood_new)
            # print("pi_new", pi_new)
            # print("alpha_new", alpha_new)
            stop=self._stopping_criteria(self.log_likelihoods,self.tol,self.max_iter)
            log_like_diff = abs(log_likelihood_new-log_likelihood_old)/abs(log_likelihood_old)
            alpha_temp = alpha_new
            pi_temp = pi_new
            if self.print_log_likelihood==True:
                print("Loglikelihood:",log_likelihood_new, "Relative Difference:", log_like_diff )
            
        return pi_new,alpha_new, log_likelihood_new,log_gamma_new
    
    def _predict(self,X):
        X=self._process_data(X)
        if self.mixture_type == "identical":
            return self._estimate_weighted_log_prob_identical(X,self.alpha_new, self.pi_new).argmax(axis=1)
        elif self.mixture_type=="nonidentical":
            return self._estimate_weighted_log_prob_nonidentical(X,self.alpha_new, self.pi_new,self.dist_comb).argmax(axis=1)
            
            
    
    def _m_step(self,gamma_matrix,X,estimate_alphas_function,dist_comb,**kwargs):
        if self.EM_type=="Hard":
            cluster,data_cwise=utils_mixture.mixture_clusters(gamma_matrix, X)
            pi_new = np.bincount(cluster, minlength=gamma_matrix.shape[1]) / len(cluster)
            alpha_new=estimate_alphas_function(data_cwise,self.alpha_not,dist_comb,**kwargs)
        elif self.EM_type=="Soft":
            nj=gamma_matrix.sum(axis=0) + 10 * np.finfo(gamma_matrix.dtype).eps
            pi_new = nj / nj.sum()
            alpha_new=estimate_alphas_function(X,gamma_matrix,self.alpha_temp,**kwargs)

        return pi_new, alpha_new

    def _stopping_criteria(self, log_likelihoods, tol, n_iter_max):
        if len(log_likelihoods) == n_iter_max:
            return True
        elif abs(log_likelihoods[-1]-log_likelihoods[-2])/abs(log_likelihoods[-2]) <= tol:
            return True
        else:
            return False
    
    def aic(self):
        """
        Compute Akaike Information Criterion (AIC).
        
        Parameters:
            log_likelihood (float): Maximum log-likelihood value from model fitting.
            num_params (int): Total number of estimated parameters in the model.
    
        Returns:
            float: AIC value.
        """
        if self.mixture_type == "nonidentical":
            return self.list_aic
        else:
            return -2 * self.log_likelihood_new + 2 * self.total_parameters


    def bic(self):
        """
        Compute Bayesian Information Criterion (BIC).
        
        Parameters:
            log_likelihood (float): Maximum log-likelihood value from model fitting.
            num_params (int): Total number of estimated parameters in the model.
            num_samples (int): Number of observations (data points).
    
        Returns:
            float: BIC value.
        """
        if self.mixture_type == "nonidentical":
            return self.list_bic
        else:
            return -2 * self.log_likelihood_new + self.total_parameters * np.log(self.n)
    
    
    def icl(self):
        """
        Compute Integrated Completed Likelihood (ICL).
        
        Parameters:
            log_likelihood (float): Maximum log-likelihood value from model fitting.
            num_params (int): Total number of estimated parameters in the model.
            num_samples (int): Number of observations (data points).
            responsibilities (numpy.ndarray): Responsibility matrix of shape (n_samples, K),
                                              where K is the number of clusters.
    
        Returns:
            float: ICL value.
        """
        if self.mixture_type == "nonidentical":
            return self.list_icl
        else:
            bic_value = self.bic()
            
            # Compute entropy term: sum over all i and k of r_ik log(r_ik)
            entropy_term = np.sum(self.gamma_temp_ar * np.log(np.clip(self.gamma_temp_ar, 1e-10, 1)))
            
            return bic_value + entropy_term
        